# SPDX-FileCopyrightText: 2017-2022 Blender Foundation
#
# SPDX-License-Identifier: GPL-2.0-or-later

import bpy

from mathutils import Color


def create(obj):  # noqa
    # generated by rigify.utils.write_metarig
    bpy.ops.object.mode_set(mode='EDIT')
    arm = obj.data

    for i in range(6):
        arm.rigify_colors.add()

    arm.rigify_colors[0].name = "Root"
    arm.rigify_colors[0].active = Color((0.5490, 1.0000, 1.0000))
    arm.rigify_colors[0].normal = Color((0.4353, 0.1843, 0.4157))
    arm.rigify_colors[0].select = Color((0.3137, 0.7843, 1.0000))
    arm.rigify_colors[0].standard_colors_lock = True
    arm.rigify_colors[1].name = "IK"
    arm.rigify_colors[1].active = Color((0.5490, 1.0000, 1.0000))
    arm.rigify_colors[1].normal = Color((0.6039, 0.0000, 0.0000))
    arm.rigify_colors[1].select = Color((0.3137, 0.7843, 1.0000))
    arm.rigify_colors[1].standard_colors_lock = True
    arm.rigify_colors[2].name = "Special"
    arm.rigify_colors[2].active = Color((0.5490, 1.0000, 1.0000))
    arm.rigify_colors[2].normal = Color((0.9569, 0.7882, 0.0471))
    arm.rigify_colors[2].select = Color((0.3137, 0.7843, 1.0000))
    arm.rigify_colors[2].standard_colors_lock = True
    arm.rigify_colors[3].name = "Tweak"
    arm.rigify_colors[3].active = Color((0.5490, 1.0000, 1.0000))
    arm.rigify_colors[3].normal = Color((0.0392, 0.2118, 0.5804))
    arm.rigify_colors[3].select = Color((0.3137, 0.7843, 1.0000))
    arm.rigify_colors[3].standard_colors_lock = True
    arm.rigify_colors[4].name = "FK"
    arm.rigify_colors[4].active = Color((0.5490, 1.0000, 1.0000))
    arm.rigify_colors[4].normal = Color((0.1176, 0.5686, 0.0353))
    arm.rigify_colors[4].select = Color((0.3137, 0.7843, 1.0000))
    arm.rigify_colors[4].standard_colors_lock = True
    arm.rigify_colors[5].name = "Extra"
    arm.rigify_colors[5].active = Color((0.5490, 1.0000, 1.0000))
    arm.rigify_colors[5].normal = Color((0.9686, 0.2510, 0.0941))
    arm.rigify_colors[5].select = Color((0.3137, 0.7843, 1.0000))
    arm.rigify_colors[5].standard_colors_lock = True

    bone_collections = {}

    for bcoll in list(arm.collections_all):
        arm.collections.remove(bcoll)

    def add_bone_collection(name, *, ui_row=0, ui_title='', sel_set=False, color_set_id=0):
        new_bcoll = arm.collections.new(name)
        new_bcoll.rigify_ui_row = ui_row
        new_bcoll.rigify_ui_title = ui_title
        new_bcoll.rigify_sel_set = sel_set
        new_bcoll.rigify_color_set_id = color_set_id
        bone_collections[name] = new_bcoll

    def assign_bone_collections(pose_bone, *coll_names):
        assert not len(pose_bone.bone.collections)
        for name in coll_names:
            bone_collections[name].assign(pose_bone)

    def assign_bone_collection_refs(params, attr_name, *coll_names):
        ref_list = getattr(params, attr_name + '_coll_refs', None)
        if ref_list is not None:
            for name in coll_names:
                ref_list.add().set_collection(bone_collections[name])

    add_bone_collection('Face', ui_row=1, color_set_id=6)
    add_bone_collection('Face (Tweak)', ui_row=2, ui_title='(Tweak)', color_set_id=4)
    add_bone_collection('Spine', ui_row=4, color_set_id=3)
    add_bone_collection('Spine (Tweak)', ui_row=5, ui_title='(Tweak)', color_set_id=4)
    add_bone_collection('Wing.L', ui_row=7, color_set_id=5)
    add_bone_collection('Wing.L (Tweak)', ui_row=8, ui_title='(Tweak)', color_set_id=4)
    add_bone_collection('Wing.R', ui_row=7, color_set_id=5)
    add_bone_collection('Wing.R (Tweak)', ui_row=8, ui_title='(Tweak)', color_set_id=4)
    add_bone_collection('Leg.L (IK)', ui_row=11, color_set_id=2)
    add_bone_collection('Leg.L (FK)', ui_row=12, ui_title='(FK)', color_set_id=5)
    add_bone_collection('Leg.L (Tweak)', ui_row=13, ui_title='(Tweak)', color_set_id=4)
    add_bone_collection('Leg.R (IK)', ui_row=11, color_set_id=2)
    add_bone_collection('Leg.R (FK)', ui_row=12, ui_title='(FK)', color_set_id=5)
    add_bone_collection('Leg.R (Tweak)', ui_row=13, ui_title='(Tweak)', color_set_id=4)
    add_bone_collection('Claws', ui_row=14, color_set_id=6)
    add_bone_collection('Claws (Tweak)', ui_row=15, ui_title='(Tweak)', color_set_id=4)
    add_bone_collection('Feathers', ui_row=9, color_set_id=6)
    add_bone_collection('Root', ui_row=18, color_set_id=1)

    bones = {}

    bone = arm.edit_bones.new('spine.003')
    bone.head = -0.0000, 0.0451, 0.0845
    bone.tail = -0.0000, 0.0192, 0.0888
    bone.roll = 0.0000
    bone.use_connect = False
    bones['spine.003'] = bone.name
    bone = arm.edit_bones.new('spine.002')
    bone.head = -0.0000, 0.0451, 0.0845
    bone.tail = -0.0000, 0.0757, 0.0880
    bone.roll = 0.0000
    bone.use_connect = False
    bone.parent = arm.edit_bones[bones['spine.003']]
    bones['spine.002'] = bone.name
    bone = arm.edit_bones.new('spine.004')
    bone.head = -0.0000, 0.0192, 0.0888
    bone.tail = -0.0000, -0.0106, 0.0979
    bone.roll = 0.0000
    bone.use_connect = True
    bone.parent = arm.edit_bones[bones['spine.003']]
    bones['spine.004'] = bone.name
    bone = arm.edit_bones.new('spine.001')
    bone.head = -0.0000, 0.0757, 0.0880
    bone.tail = -0.0000, 0.1039, 0.0907
    bone.roll = 0.0000
    bone.use_connect = True
    bone.parent = arm.edit_bones[bones['spine.002']]
    bones['spine.001'] = bone.name
    bone = arm.edit_bones.new('spine.005')
    bone.head = -0.0000, -0.0106, 0.0979
    bone.tail = -0.0000, -0.0298, 0.1158
    bone.roll = 0.0000
    bone.use_connect = True
    bone.parent = arm.edit_bones[bones['spine.004']]
    bones['spine.005'] = bone.name
    bone = arm.edit_bones.new('pelvis.L')
    bone.head = -0.0000, 0.0192, 0.0888
    bone.tail = 0.0250, 0.0070, 0.0782
    bone.roll = -0.0369
    bone.use_connect = False
    bone.parent = arm.edit_bones[bones['spine.004']]
    bones['pelvis.L'] = bone.name
    bone = arm.edit_bones.new('pelvis.R')
    bone.head = 0.0000, 0.0192, 0.0888
    bone.tail = -0.0250, 0.0070, 0.0782
    bone.roll = 0.0369
    bone.use_connect = False
    bone.parent = arm.edit_bones[bones['spine.004']]
    bones['pelvis.R'] = bone.name
    bone = arm.edit_bones.new('thigh.L')
    bone.head = 0.0149, 0.0063, 0.0739
    bone.tail = 0.0149, 0.0143, 0.0391
    bone.roll = 0.0000
    bone.use_connect = False
    bone.parent = arm.edit_bones[bones['spine.004']]
    bones['thigh.L'] = bone.name
    bone = arm.edit_bones.new('thigh.R')
    bone.head = -0.0149, 0.0063, 0.0739
    bone.tail = -0.0149, 0.0143, 0.0391
    bone.roll = -0.0000
    bone.use_connect = False
    bone.parent = arm.edit_bones[bones['spine.004']]
    bones['thigh.R'] = bone.name
    bone = arm.edit_bones.new('spine')
    bone.head = -0.0000, 0.1039, 0.0907
    bone.tail = -0.0000, 0.1371, 0.0894
    bone.roll = 0.0000
    bone.use_connect = True
    bone.parent = arm.edit_bones[bones['spine.001']]
    bones['spine'] = bone.name
    bone = arm.edit_bones.new('shoulder.L')
    bone.head = 0.0014, -0.0217, 0.0893
    bone.tail = 0.0076, -0.0020, 0.1179
    bone.roll = 1.3977
    bone.use_connect = False
    bone.parent = arm.edit_bones[bones['spine.005']]
    bones['shoulder.L'] = bone.name
    bone = arm.edit_bones.new('spine.006')
    bone.head = -0.0000, -0.0298, 0.1158
    bone.tail = -0.0000, -0.0417, 0.1348
    bone.roll = 0.0000
    bone.use_connect = True
    bone.parent = arm.edit_bones[bones['spine.005']]
    bones['spine.006'] = bone.name
    bone = arm.edit_bones.new('shoulder.R')
    bone.head = -0.0014, -0.0217, 0.0893
    bone.tail = -0.0076, -0.0020, 0.1179
    bone.roll = -1.3977
    bone.use_connect = False
    bone.parent = arm.edit_bones[bones['spine.005']]
    bones['shoulder.R'] = bone.name
    bone = arm.edit_bones.new('shin.L')
    bone.head = 0.0149, 0.0143, 0.0391
    bone.tail = 0.0149, 0.0015, 0.0074
    bone.roll = 0.0000
    bone.use_connect = True
    bone.parent = arm.edit_bones[bones['thigh.L']]
    bones['shin.L'] = bone.name
    bone = arm.edit_bones.new('shin.R')
    bone.head = -0.0149, 0.0143, 0.0391
    bone.tail = -0.0149, 0.0015, 0.0074
    bone.roll = -0.0000
    bone.use_connect = True
    bone.parent = arm.edit_bones[bones['thigh.R']]
    bones['shin.R'] = bone.name
    bone = arm.edit_bones.new('t_feather.L')
    bone.head = 0.0112, 0.1017, 0.0907
    bone.tail = 0.0167, 0.1345, 0.0894
    bone.roll = 0.0032
    bone.use_connect = False
    bone.parent = arm.edit_bones[bones['spine']]
    bones['t_feather.L'] = bone.name
    bone = arm.edit_bones.new('t_feather.R')
    bone.head = -0.0112, 0.1017, 0.0907
    bone.tail = -0.0167, 0.1345, 0.0894
    bone.roll = -0.0032
    bone.use_connect = False
    bone.parent = arm.edit_bones[bones['spine']]
    bones['t_feather.R'] = bone.name
    bone = arm.edit_bones.new('Wing.L')
    bone.head = 0.0089, 0.0141, 0.1157
    bone.tail = 0.0485, 0.0107, 0.1163
    bone.roll = 2.8221
    bone.use_connect = False
    bone.parent = arm.edit_bones[bones['shoulder.L']]
    bones['Wing.L'] = bone.name
    bone = arm.edit_bones.new('neck.001')
    bone.head = -0.0000, -0.0417, 0.1348
    bone.tail = -0.0000, -0.0458, 0.1429
    bone.roll = 0.0001
    bone.use_connect = False
    bone.parent = arm.edit_bones[bones['spine.006']]
    bones['neck.001'] = bone.name
    bone = arm.edit_bones.new('Wing.R')
    bone.head = -0.0089, 0.0141, 0.1157
    bone.tail = -0.0485, 0.0107, 0.1163
    bone.roll = -2.8221
    bone.use_connect = False
    bone.parent = arm.edit_bones[bones['shoulder.R']]
    bones['Wing.R'] = bone.name
    bone = arm.edit_bones.new('foot.L')
    bone.head = 0.0149, 0.0015, 0.0074
    bone.tail = 0.0149, -0.0033, 0.0045
    bone.roll = 0.0000
    bone.use_connect = True
    bone.parent = arm.edit_bones[bones['shin.L']]
    bones['foot.L'] = bone.name
    bone = arm.edit_bones.new('foot.R')
    bone.head = -0.0149, 0.0015, 0.0074
    bone.tail = -0.0149, -0.0033, 0.0045
    bone.roll = -0.0000
    bone.use_connect = True
    bone.parent = arm.edit_bones[bones['shin.R']]
    bones['foot.R'] = bone.name
    bone = arm.edit_bones.new('Wing.001.L')
    bone.head = 0.0485, 0.0107, 0.1163
    bone.tail = 0.0919, 0.0021, 0.1191
    bone.roll = 2.7596
    bone.use_connect = True
    bone.parent = arm.edit_bones[bones['Wing.L']]
    bones['Wing.001.L'] = bone.name
    bone = arm.edit_bones.new('w_feather.004.L')
    bone.head = 0.0382, 0.0253, 0.1081
    bone.tail = 0.0553, 0.0878, 0.0901
    bone.roll = 3.0872
    bone.use_connect = False
    bone.parent = arm.edit_bones[bones['Wing.L']]
    bones['w_feather.004.L'] = bone.name
    bone = arm.edit_bones.new('neck.002')
    bone.head = -0.0000, -0.0458, 0.1429
    bone.tail = -0.0000, -0.0483, 0.1498
    bone.roll = 0.0001
    bone.use_connect = True
    bone.parent = arm.edit_bones[bones['neck.001']]
    bones['neck.002'] = bone.name
    bone = arm.edit_bones.new('Wing.001.R')
    bone.head = -0.0485, 0.0107, 0.1163
    bone.tail = -0.0919, 0.0021, 0.1191
    bone.roll = -2.7596
    bone.use_connect = True
    bone.parent = arm.edit_bones[bones['Wing.R']]
    bones['Wing.001.R'] = bone.name
    bone = arm.edit_bones.new('w_feather.004.R')
    bone.head = -0.0382, 0.0253, 0.1081
    bone.tail = -0.0553, 0.0878, 0.0901
    bone.roll = -3.0872
    bone.use_connect = False
    bone.parent = arm.edit_bones[bones['Wing.R']]
    bones['w_feather.004.R'] = bone.name
    bone = arm.edit_bones.new('toe.L')
    bone.head = 0.0149, -0.0033, 0.0045
    bone.tail = 0.0149, -0.0404, 0.0006
    bone.roll = 3.1416
    bone.use_connect = True
    bone.parent = arm.edit_bones[bones['foot.L']]
    bones['toe.L'] = bone.name
    bone = arm.edit_bones.new('toe.R')
    bone.head = -0.0149, -0.0033, 0.0045
    bone.tail = -0.0149, -0.0404, 0.0006
    bone.roll = -3.1416
    bone.use_connect = True
    bone.parent = arm.edit_bones[bones['foot.R']]
    bones['toe.R'] = bone.name
    bone = arm.edit_bones.new('Wing.002.L')
    bone.head = 0.0919, 0.0021, 0.1191
    bone.tail = 0.2128, 0.0178, 0.1159
    bone.roll = 2.8565
    bone.use_connect = True
    bone.parent = arm.edit_bones[bones['Wing.001.L']]
    bones['Wing.002.L'] = bone.name
    bone = arm.edit_bones.new('w_feather.003.L')
    bone.head = 0.0754, 0.0212, 0.1122
    bone.tail = 0.1342, 0.0830, 0.0901
    bone.roll = 3.0993
    bone.use_connect = False
    bone.parent = arm.edit_bones[bones['Wing.001.L']]
    bones['w_feather.003.L'] = bone.name
    bone = arm.edit_bones.new('head')
    bone.head = -0.0000, -0.0483, 0.1498
    bone.tail = -0.0000, -0.1026, 0.1815
    bone.roll = 0.0000
    bone.use_connect = True
    bone.parent = arm.edit_bones[bones['neck.002']]
    bones['head'] = bone.name
    bone = arm.edit_bones.new('Wing.002.R')
    bone.head = -0.0919, 0.0021, 0.1191
    bone.tail = -0.2128, 0.0178, 0.1159
    bone.roll = -2.8565
    bone.use_connect = True
    bone.parent = arm.edit_bones[bones['Wing.001.R']]
    bones['Wing.002.R'] = bone.name
    bone = arm.edit_bones.new('w_feather.003.R')
    bone.head = -0.0754, 0.0212, 0.1122
    bone.tail = -0.1342, 0.0830, 0.0901
    bone.roll = -3.0993
    bone.use_connect = False
    bone.parent = arm.edit_bones[bones['Wing.001.R']]
    bones['w_feather.003.R'] = bone.name
    bone = arm.edit_bones.new('toes_parent.L')
    bone.head = 0.0149, -0.0033, 0.0045
    bone.tail = 0.0149, -0.0096, 0.0037
    bone.roll = 3.1416
    bone.use_connect = False
    bone.parent = arm.edit_bones[bones['toe.L']]
    bones['toes_parent.L'] = bone.name
    bone = arm.edit_bones.new('t_thumb.001.L')
    bone.head = 0.0136, 0.0051, 0.0022
    bone.tail = 0.0131, 0.0142, 0.0026
    bone.roll = 3.1406
    bone.use_connect = False
    bone.parent = arm.edit_bones[bones['toe.L']]
    bones['t_thumb.001.L'] = bone.name
    bone = arm.edit_bones.new('toes_parent.R')
    bone.head = -0.0149, -0.0033, 0.0045
    bone.tail = -0.0149, -0.0096, 0.0037
    bone.roll = -3.1416
    bone.use_connect = False
    bone.parent = arm.edit_bones[bones['toe.R']]
    bones['toes_parent.R'] = bone.name
    bone = arm.edit_bones.new('t_thumb.001.R')
    bone.head = -0.0136, 0.0051, 0.0022
    bone.tail = -0.0131, 0.0142, 0.0026
    bone.roll = -3.1406
    bone.use_connect = False
    bone.parent = arm.edit_bones[bones['toe.R']]
    bones['t_thumb.001.R'] = bone.name
    bone = arm.edit_bones.new('w_feather.001.L')
    bone.head = 0.1595, -0.0062, 0.1163
    bone.tail = 0.2489, -0.0055, 0.1224
    bone.roll = 2.9290
    bone.use_connect = False
    bone.parent = arm.edit_bones[bones['Wing.002.L']]
    bones['w_feather.001.L'] = bone.name
    bone = arm.edit_bones.new('w_feather.002.L')
    bone.head = 0.1218, 0.0331, 0.1099
    bone.tail = 0.1812, 0.0495, 0.1068
    bone.roll = 2.8770
    bone.use_connect = False
    bone.parent = arm.edit_bones[bones['Wing.002.L']]
    bones['w_feather.002.L'] = bone.name
    bone = arm.edit_bones.new('beak.001.T')
    bone.head = -0.0000, -0.0703, 0.1580
    bone.tail = -0.0000, -0.0927, 0.1597
    bone.roll = 0.0001
    bone.use_connect = False
    bone.parent = arm.edit_bones[bones['head']]
    bones['beak.001.T'] = bone.name
    bone = arm.edit_bones.new('beak_001.B')
    bone.head = -0.0000, -0.0703, 0.1556
    bone.tail = -0.0000, -0.0914, 0.1555
    bone.roll = 0.0001
    bone.use_connect = False
    bone.parent = arm.edit_bones[bones['head']]
    bones['beak_001.B'] = bone.name
    bone = arm.edit_bones.new('eye.L')
    bone.head = 0.0055, -0.0647, 0.1615
    bone.tail = 0.0198, -0.0675, 0.1615
    bone.roll = 0.0000
    bone.use_connect = False
    bone.parent = arm.edit_bones[bones['head']]
    bones['eye.L'] = bone.name
    bone = arm.edit_bones.new('eye.R')
    bone.head = -0.0055, -0.0647, 0.1615
    bone.tail = -0.0198, -0.0675, 0.1615
    bone.roll = -0.0000
    bone.use_connect = False
    bone.parent = arm.edit_bones[bones['head']]
    bones['eye.R'] = bone.name
    bone = arm.edit_bones.new('n_feather.001.L')
    bone.head = 0.0098, -0.0390, 0.1566
    bone.tail = 0.0131, -0.0247, 0.1430
    bone.roll = -0.0380
    bone.use_connect = False
    bone.parent = arm.edit_bones[bones['head']]
    bones['n_feather.001.L'] = bone.name
    bone = arm.edit_bones.new('n_feather.001.R')
    bone.head = -0.0098, -0.0390, 0.1566
    bone.tail = -0.0131, -0.0247, 0.1430
    bone.roll = 0.0380
    bone.use_connect = False
    bone.parent = arm.edit_bones[bones['head']]
    bones['n_feather.001.R'] = bone.name
    bone = arm.edit_bones.new('skull.006.L')
    bone.head = 0.0033, -0.0587, 0.1538
    bone.tail = 0.0162, -0.0666, 0.1509
    bone.roll = 1.6532
    bone.use_connect = False
    bone.parent = arm.edit_bones[bones['head']]
    bones['skull.006.L'] = bone.name
    bone = arm.edit_bones.new('skull.006.R')
    bone.head = -0.0033, -0.0587, 0.1538
    bone.tail = -0.0162, -0.0666, 0.1509
    bone.roll = -1.6532
    bone.use_connect = False
    bone.parent = arm.edit_bones[bones['head']]
    bones['skull.006.R'] = bone.name
    bone = arm.edit_bones.new('w_feather.001.R')
    bone.head = -0.1595, -0.0062, 0.1163
    bone.tail = -0.2489, -0.0055, 0.1224
    bone.roll = -2.9290
    bone.use_connect = False
    bone.parent = arm.edit_bones[bones['Wing.002.R']]
    bones['w_feather.001.R'] = bone.name
    bone = arm.edit_bones.new('w_feather.002.R')
    bone.head = -0.1218, 0.0331, 0.1099
    bone.tail = -0.1812, 0.0495, 0.1068
    bone.roll = -2.8770
    bone.use_connect = False
    bone.parent = arm.edit_bones[bones['Wing.002.R']]
    bones['w_feather.002.R'] = bone.name
    bone = arm.edit_bones.new('t_ring.001.L')
    bone.head = 0.0183, -0.0026, 0.0034
    bone.tail = 0.0216, -0.0129, 0.0038
    bone.roll = 6.0340
    bone.use_connect = False
    bone.parent = arm.edit_bones[bones['toes_parent.L']]
    bones['t_ring.001.L'] = bone.name
    bone = arm.edit_bones.new('t_index.001.L')
    bone.head = 0.0122, -0.0026, 0.0034
    bone.tail = 0.0093, -0.0130, 0.0038
    bone.roll = 6.5631
    bone.use_connect = False
    bone.parent = arm.edit_bones[bones['toes_parent.L']]
    bones['t_index.001.L'] = bone.name
    bone = arm.edit_bones.new('t_middle.001.L')
    bone.head = 0.0149, -0.0050, 0.0034
    bone.tail = 0.0149, -0.0207, 0.0040
    bone.roll = 3.1416
    bone.use_connect = False
    bone.parent = arm.edit_bones[bones['toes_parent.L']]
    bones['t_middle.001.L'] = bone.name
    bone = arm.edit_bones.new('t_thumb.002.L')
    bone.head = 0.0131, 0.0142, 0.0026
    bone.tail = 0.0126, 0.0241, 0.0003
    bone.roll = 3.1472
    bone.use_connect = True
    bone.parent = arm.edit_bones[bones['t_thumb.001.L']]
    bones['t_thumb.002.L'] = bone.name
    bone = arm.edit_bones.new('t_ring.001.R')
    bone.head = -0.0183, -0.0026, 0.0034
    bone.tail = -0.0216, -0.0129, 0.0038
    bone.roll = -6.0340
    bone.use_connect = False
    bone.parent = arm.edit_bones[bones['toes_parent.R']]
    bones['t_ring.001.R'] = bone.name
    bone = arm.edit_bones.new('t_index.001.R')
    bone.head = -0.0122, -0.0026, 0.0034
    bone.tail = -0.0093, -0.0130, 0.0038
    bone.roll = -6.5631
    bone.use_connect = False
    bone.parent = arm.edit_bones[bones['toes_parent.R']]
    bones['t_index.001.R'] = bone.name
    bone = arm.edit_bones.new('t_middle.001.R')
    bone.head = -0.0149, -0.0050, 0.0034
    bone.tail = -0.0149, -0.0207, 0.0040
    bone.roll = -3.1416
    bone.use_connect = False
    bone.parent = arm.edit_bones[bones['toes_parent.R']]
    bones['t_middle.001.R'] = bone.name
    bone = arm.edit_bones.new('t_thumb.002.R')
    bone.head = -0.0131, 0.0142, 0.0026
    bone.tail = -0.0126, 0.0241, 0.0003
    bone.roll = -3.1472
    bone.use_connect = True
    bone.parent = arm.edit_bones[bones['t_thumb.001.R']]
    bones['t_thumb.002.R'] = bone.name
    bone = arm.edit_bones.new('beak.002.T')
    bone.head = -0.0000, -0.0927, 0.1597
    bone.tail = -0.0000, -0.1131, 0.1533
    bone.roll = 0.0001
    bone.use_connect = True
    bone.parent = arm.edit_bones[bones['beak.001.T']]
    bones['beak.002.T'] = bone.name
    bone = arm.edit_bones.new('beak.002.B')
    bone.head = -0.0000, -0.0914, 0.1555
    bone.tail = -0.0000, -0.1084, 0.1519
    bone.roll = 0.0001
    bone.use_connect = True
    bone.parent = arm.edit_bones[bones['beak_001.B']]
    bones['beak.002.B'] = bone.name
    bone = arm.edit_bones.new('tongue.001.L')
    bone.head = -0.0000, -0.0792, 0.1536
    bone.tail = -0.0000, -0.0840, 0.1555
    bone.roll = 0.0001
    bone.use_connect = False
    bone.parent = arm.edit_bones[bones['beak_001.B']]
    bones['tongue.001.L'] = bone.name
    bone = arm.edit_bones.new('t_ring.002.L')
    bone.head = 0.0216, -0.0129, 0.0038
    bone.tail = 0.0237, -0.0195, 0.0032
    bone.roll = 0.5806
    bone.use_connect = True
    bone.parent = arm.edit_bones[bones['t_ring.001.L']]
    bones['t_ring.002.L'] = bone.name
    bone = arm.edit_bones.new('t_index.002.L')
    bone.head = 0.0093, -0.0130, 0.0038
    bone.tail = 0.0067, -0.0223, 0.0032
    bone.roll = -0.4957
    bone.use_connect = True
    bone.parent = arm.edit_bones[bones['t_index.001.L']]
    bones['t_index.002.L'] = bone.name
    bone = arm.edit_bones.new('t_middle.002.L')
    bone.head = 0.0149, -0.0207, 0.0040
    bone.tail = 0.0149, -0.0308, 0.0031
    bone.roll = 3.1416
    bone.use_connect = True
    bone.parent = arm.edit_bones[bones['t_middle.001.L']]
    bones['t_middle.002.L'] = bone.name
    bone = arm.edit_bones.new('t_ring.002.R')
    bone.head = -0.0216, -0.0129, 0.0038
    bone.tail = -0.0237, -0.0195, 0.0032
    bone.roll = -0.5806
    bone.use_connect = True
    bone.parent = arm.edit_bones[bones['t_ring.001.R']]
    bones['t_ring.002.R'] = bone.name
    bone = arm.edit_bones.new('t_index.002.R')
    bone.head = -0.0093, -0.0130, 0.0038
    bone.tail = -0.0067, -0.0223, 0.0032
    bone.roll = 0.4957
    bone.use_connect = True
    bone.parent = arm.edit_bones[bones['t_index.001.R']]
    bones['t_index.002.R'] = bone.name
    bone = arm.edit_bones.new('t_middle.002.R')
    bone.head = -0.0149, -0.0207, 0.0040
    bone.tail = -0.0149, -0.0308, 0.0031
    bone.roll = -3.1416
    bone.use_connect = True
    bone.parent = arm.edit_bones[bones['t_middle.001.R']]
    bones['t_middle.002.R'] = bone.name
    bone = arm.edit_bones.new('tongue.002.L')
    bone.head = -0.0000, -0.0840, 0.1555
    bone.tail = -0.0000, -0.0874, 0.1553
    bone.roll = 0.0001
    bone.use_connect = True
    bone.parent = arm.edit_bones[bones['tongue.001.L']]
    bones['tongue.002.L'] = bone.name
    bone = arm.edit_bones.new('t_ring.003.L')
    bone.head = 0.0237, -0.0195, 0.0032
    bone.tail = 0.0261, -0.0270, 0.0009
    bone.roll = 1.5954
    bone.use_connect = True
    bone.parent = arm.edit_bones[bones['t_ring.002.L']]
    bones['t_ring.003.L'] = bone.name
    bone = arm.edit_bones.new('t_index.003.L')
    bone.head = 0.0067, -0.0223, 0.0032
    bone.tail = 0.0038, -0.0327, 0.0009
    bone.roll = 4.8853
    bone.use_connect = True
    bone.parent = arm.edit_bones[bones['t_index.002.L']]
    bones['t_index.003.L'] = bone.name
    bone = arm.edit_bones.new('t_middle.003.L')
    bone.head = 0.0149, -0.0308, 0.0031
    bone.tail = 0.0149, -0.0421, -0.0003
    bone.roll = 3.1416
    bone.use_connect = True
    bone.parent = arm.edit_bones[bones['t_middle.002.L']]
    bones['t_middle.003.L'] = bone.name
    bone = arm.edit_bones.new('t_ring.003.R')
    bone.head = -0.0237, -0.0195, 0.0032
    bone.tail = -0.0261, -0.0270, 0.0009
    bone.roll = -1.5954
    bone.use_connect = True
    bone.parent = arm.edit_bones[bones['t_ring.002.R']]
    bones['t_ring.003.R'] = bone.name
    bone = arm.edit_bones.new('t_index.003.R')
    bone.head = -0.0067, -0.0223, 0.0032
    bone.tail = -0.0038, -0.0327, 0.0009
    bone.roll = -4.8853
    bone.use_connect = True
    bone.parent = arm.edit_bones[bones['t_index.002.R']]
    bones['t_index.003.R'] = bone.name
    bone = arm.edit_bones.new('t_middle.003.R')
    bone.head = -0.0149, -0.0308, 0.0031
    bone.tail = -0.0149, -0.0421, -0.0003
    bone.roll = -3.1416
    bone.use_connect = True
    bone.parent = arm.edit_bones[bones['t_middle.002.R']]
    bones['t_middle.003.R'] = bone.name
    bone = arm.edit_bones.new('tongue.003.L')
    bone.head = -0.0000, -0.0874, 0.1553
    bone.tail = -0.0000, -0.0898, 0.1553
    bone.roll = 3.1416
    bone.use_connect = True
    bone.parent = arm.edit_bones[bones['tongue.002.L']]
    bones['tongue.003.L'] = bone.name

    bpy.ops.object.mode_set(mode='OBJECT')
    pbone = obj.pose.bones[bones['spine.003']]
    pbone.rigify_type = 'spines.basic_spine'
    pbone.lock_location = (False, False, False)
    pbone.lock_rotation = (False, False, False)
    pbone.lock_rotation_w = False
    pbone.lock_scale = (False, False, False)
    pbone.rotation_mode = 'QUATERNION'
    assign_bone_collections(pbone, 'Spine')
    try:
        pbone.rigify_parameters.pivot_pos = 1
    except AttributeError:
        pass
    assign_bone_collection_refs(pbone.rigify_parameters, 'tweak', 'Spine (Tweak)')
    assign_bone_collection_refs(pbone.rigify_parameters, 'fk', 'Spine (Tweak)')
    pbone = obj.pose.bones[bones['spine.002']]
    pbone.rigify_type = 'spines.basic_tail'
    pbone.lock_location = (False, False, False)
    pbone.lock_rotation = (False, False, False)
    pbone.lock_rotation_w = False
    pbone.lock_scale = (False, False, False)
    pbone.rotation_mode = 'QUATERNION'
    assign_bone_collections(pbone, 'Spine')
    try:
        pbone.rigify_parameters.copy_rotation_axes = (True, False, True)
    except AttributeError:
        pass
    try:
        pbone.rigify_parameters.connect_chain = True
    except AttributeError:
        pass
    assign_bone_collection_refs(pbone.rigify_parameters, 'tweak', 'Spine (Tweak)')
    pbone = obj.pose.bones[bones['spine.004']]
    pbone.rigify_type = ''
    pbone.lock_location = (False, False, False)
    pbone.lock_rotation = (False, False, False)
    pbone.lock_rotation_w = False
    pbone.lock_scale = (False, False, False)
    pbone.rotation_mode = 'QUATERNION'
    assign_bone_collections(pbone, 'Spine')
    pbone = obj.pose.bones[bones['spine.001']]
    pbone.rigify_type = ''
    pbone.lock_location = (False, False, False)
    pbone.lock_rotation = (False, False, False)
    pbone.lock_rotation_w = False
    pbone.lock_scale = (False, False, False)
    pbone.rotation_mode = 'QUATERNION'
    assign_bone_collections(pbone, 'Spine')
    pbone = obj.pose.bones[bones['spine.005']]
    pbone.rigify_type = ''
    pbone.lock_location = (False, False, False)
    pbone.lock_rotation = (False, False, False)
    pbone.lock_rotation_w = False
    pbone.lock_scale = (False, False, False)
    pbone.rotation_mode = 'QUATERNION'
    assign_bone_collections(pbone, 'Spine')
    pbone = obj.pose.bones[bones['pelvis.L']]
    pbone.rigify_type = 'basic.super_copy'
    pbone.lock_location = (False, False, False)
    pbone.lock_rotation = (False, False, False)
    pbone.lock_rotation_w = False
    pbone.lock_scale = (False, False, False)
    pbone.rotation_mode = 'QUATERNION'
    assign_bone_collections(pbone, 'Spine')
    try:
        pbone.rigify_parameters.make_widget = False
    except AttributeError:
        pass
    try:
        pbone.rigify_parameters.make_control = False
    except AttributeError:
        pass
    pbone = obj.pose.bones[bones['pelvis.R']]
    pbone.rigify_type = 'basic.super_copy'
    pbone.lock_location = (False, False, False)
    pbone.lock_rotation = (False, False, False)
    pbone.lock_rotation_w = False
    pbone.lock_scale = (False, False, False)
    pbone.rotation_mode = 'QUATERNION'
    assign_bone_collections(pbone, 'Spine')
    try:
        pbone.rigify_parameters.make_control = False
    except AttributeError:
        pass
    try:
        pbone.rigify_parameters.make_widget = False
    except AttributeError:
        pass
    pbone = obj.pose.bones[bones['thigh.L']]
    pbone.rigify_type = 'limbs.paw'
    pbone.lock_location = (False, False, False)
    pbone.lock_rotation = (False, False, False)
    pbone.lock_rotation_w = False
    pbone.lock_scale = (False, False, False)
    pbone.rotation_mode = 'QUATERNION'
    assign_bone_collections(pbone, 'Leg.L (IK)')
    try:
        pbone.rigify_parameters.limb_type = 'paw'
    except AttributeError:
        pass
    try:
        pbone.rigify_parameters.ik_local_location = False
    except AttributeError:
        pass
    assign_bone_collection_refs(pbone.rigify_parameters, 'tweak', 'Leg.L (Tweak)')
    assign_bone_collection_refs(pbone.rigify_parameters, 'fk', 'Leg.L (FK)')
    pbone = obj.pose.bones[bones['thigh.R']]
    pbone.rigify_type = 'limbs.paw'
    pbone.lock_location = (False, False, False)
    pbone.lock_rotation = (False, False, False)
    pbone.lock_rotation_w = False
    pbone.lock_scale = (False, False, False)
    pbone.rotation_mode = 'QUATERNION'
    assign_bone_collections(pbone, 'Leg.R (IK)')
    try:
        pbone.rigify_parameters.limb_type = 'paw'
    except AttributeError:
        pass
    try:
        pbone.rigify_parameters.ik_local_location = False
    except AttributeError:
        pass
    assign_bone_collection_refs(pbone.rigify_parameters, 'tweak', 'Leg.R (Tweak)')
    assign_bone_collection_refs(pbone.rigify_parameters, 'fk', 'Leg.R (FK)')
    pbone = obj.pose.bones[bones['spine']]
    pbone.rigify_type = ''
    pbone.lock_location = (False, False, False)
    pbone.lock_rotation = (False, False, False)
    pbone.lock_rotation_w = False
    pbone.lock_scale = (False, False, False)
    pbone.rotation_mode = 'QUATERNION'
    assign_bone_collections(pbone, 'Spine')
    pbone = obj.pose.bones[bones['shoulder.L']]
    pbone.rigify_type = 'basic.super_copy'
    pbone.lock_location = (False, False, False)
    pbone.lock_rotation = (False, False, False)
    pbone.lock_rotation_w = False
    pbone.lock_scale = (False, False, False)
    pbone.rotation_mode = 'QUATERNION'
    assign_bone_collections(pbone, 'Spine')
    try:
        pbone.rigify_parameters.make_widget = False
    except AttributeError:
        pass
    pbone = obj.pose.bones[bones['spine.006']]
    pbone.rigify_type = ''
    pbone.lock_location = (False, False, False)
    pbone.lock_rotation = (False, False, False)
    pbone.lock_rotation_w = False
    pbone.lock_scale = (False, False, False)
    pbone.rotation_mode = 'QUATERNION'
    assign_bone_collections(pbone, 'Spine')
    pbone = obj.pose.bones[bones['shoulder.R']]
    pbone.rigify_type = 'basic.super_copy'
    pbone.lock_location = (False, False, False)
    pbone.lock_rotation = (False, False, False)
    pbone.lock_rotation_w = False
    pbone.lock_scale = (False, False, False)
    pbone.rotation_mode = 'QUATERNION'
    assign_bone_collections(pbone, 'Spine')
    try:
        pbone.rigify_parameters.make_widget = False
    except AttributeError:
        pass
    pbone = obj.pose.bones[bones['shin.L']]
    pbone.rigify_type = ''
    pbone.lock_location = (False, False, False)
    pbone.lock_rotation = (False, False, False)
    pbone.lock_rotation_w = False
    pbone.lock_scale = (False, False, False)
    pbone.rotation_mode = 'QUATERNION'
    assign_bone_collections(pbone, 'Leg.L (IK)')
    pbone = obj.pose.bones[bones['shin.R']]
    pbone.rigify_type = ''
    pbone.lock_location = (False, False, False)
    pbone.lock_rotation = (False, False, False)
    pbone.lock_rotation_w = False
    pbone.lock_scale = (False, False, False)
    pbone.rotation_mode = 'QUATERNION'
    assign_bone_collections(pbone, 'Leg.R (IK)')
    pbone = obj.pose.bones[bones['t_feather.L']]
    pbone.rigify_type = 'basic.super_copy'
    pbone.lock_location = (False, False, False)
    pbone.lock_rotation = (False, False, False)
    pbone.lock_rotation_w = False
    pbone.lock_scale = (False, False, False)
    pbone.rotation_mode = 'QUATERNION'
    assign_bone_collections(pbone, 'Feathers')
    try:
        pbone.rigify_parameters.make_widget = False
    except AttributeError:
        pass
    pbone = obj.pose.bones[bones['t_feather.R']]
    pbone.rigify_type = 'basic.super_copy'
    pbone.lock_location = (False, False, False)
    pbone.lock_rotation = (False, False, False)
    pbone.lock_rotation_w = False
    pbone.lock_scale = (False, False, False)
    pbone.rotation_mode = 'QUATERNION'
    assign_bone_collections(pbone, 'Feathers')
    try:
        pbone.rigify_parameters.make_widget = False
    except AttributeError:
        pass
    pbone = obj.pose.bones[bones['Wing.L']]
    pbone.rigify_type = 'limbs.simple_tentacle'
    pbone.lock_location = (False, False, False)
    pbone.lock_rotation = (False, False, False)
    pbone.lock_rotation_w = False
    pbone.lock_scale = (False, False, False)
    pbone.rotation_mode = 'QUATERNION'
    assign_bone_collections(pbone, 'Wing.L')
    try:
        pbone.rigify_parameters.copy_rotation_axes = (False, False, False)
    except AttributeError:
        pass
    assign_bone_collection_refs(pbone.rigify_parameters, 'tweak', 'Wing.L (Tweak)')
    pbone = obj.pose.bones[bones['neck.001']]
    pbone.rigify_type = 'spines.super_head'
    pbone.lock_location = (False, False, False)
    pbone.lock_rotation = (False, False, False)
    pbone.lock_rotation_w = False
    pbone.lock_scale = (False, False, False)
    pbone.rotation_mode = 'QUATERNION'
    assign_bone_collections(pbone, 'Spine')
    try:
        pbone.rigify_parameters.connect_chain = True
    except AttributeError:
        pass
    assign_bone_collection_refs(pbone.rigify_parameters, 'tweak', 'Spine (Tweak)')
    pbone = obj.pose.bones[bones['Wing.R']]
    pbone.rigify_type = 'limbs.simple_tentacle'
    pbone.lock_location = (False, False, False)
    pbone.lock_rotation = (False, False, False)
    pbone.lock_rotation_w = False
    pbone.lock_scale = (False, False, False)
    pbone.rotation_mode = 'QUATERNION'
    assign_bone_collections(pbone, 'Wing.R')
    try:
        pbone.rigify_parameters.copy_rotation_axes = (False, False, False)
    except AttributeError:
        pass
    assign_bone_collection_refs(pbone.rigify_parameters, 'tweak', 'Wing.R (Tweak)')
    pbone = obj.pose.bones[bones['foot.L']]
    pbone.rigify_type = ''
    pbone.lock_location = (False, False, False)
    pbone.lock_rotation = (False, False, False)
    pbone.lock_rotation_w = False
    pbone.lock_scale = (False, False, False)
    pbone.rotation_mode = 'QUATERNION'
    assign_bone_collections(pbone, 'Leg.L (IK)')
    pbone = obj.pose.bones[bones['foot.R']]
    pbone.rigify_type = ''
    pbone.lock_location = (False, False, False)
    pbone.lock_rotation = (False, False, False)
    pbone.lock_rotation_w = False
    pbone.lock_scale = (False, False, False)
    pbone.rotation_mode = 'QUATERNION'
    assign_bone_collections(pbone, 'Leg.R (IK)')
    pbone = obj.pose.bones[bones['Wing.001.L']]
    pbone.rigify_type = ''
    pbone.lock_location = (False, False, False)
    pbone.lock_rotation = (False, False, False)
    pbone.lock_rotation_w = False
    pbone.lock_scale = (False, False, False)
    pbone.rotation_mode = 'QUATERNION'
    assign_bone_collections(pbone, 'Wing.L')
    pbone = obj.pose.bones[bones['w_feather.004.L']]
    pbone.rigify_type = 'basic.super_copy'
    pbone.lock_location = (False, False, False)
    pbone.lock_rotation = (False, False, False)
    pbone.lock_rotation_w = False
    pbone.lock_scale = (False, False, False)
    pbone.rotation_mode = 'QUATERNION'
    assign_bone_collections(pbone, 'Feathers')
    try:
        pbone.rigify_parameters.make_widget = False
    except AttributeError:
        pass
    pbone = obj.pose.bones[bones['neck.002']]
    pbone.rigify_type = ''
    pbone.lock_location = (False, False, False)
    pbone.lock_rotation = (False, False, False)
    pbone.lock_rotation_w = False
    pbone.lock_scale = (False, False, False)
    pbone.rotation_mode = 'QUATERNION'
    assign_bone_collections(pbone, 'Spine')
    pbone = obj.pose.bones[bones['Wing.001.R']]
    pbone.rigify_type = ''
    pbone.lock_location = (False, False, False)
    pbone.lock_rotation = (False, False, False)
    pbone.lock_rotation_w = False
    pbone.lock_scale = (False, False, False)
    pbone.rotation_mode = 'QUATERNION'
    assign_bone_collections(pbone, 'Wing.R')
    pbone = obj.pose.bones[bones['w_feather.004.R']]
    pbone.rigify_type = 'basic.super_copy'
    pbone.lock_location = (False, False, False)
    pbone.lock_rotation = (False, False, False)
    pbone.lock_rotation_w = False
    pbone.lock_scale = (False, False, False)
    pbone.rotation_mode = 'QUATERNION'
    assign_bone_collections(pbone, 'Feathers')
    try:
        pbone.rigify_parameters.make_widget = False
    except AttributeError:
        pass
    pbone = obj.pose.bones[bones['toe.L']]
    pbone.rigify_type = ''
    pbone.lock_location = (False, False, False)
    pbone.lock_rotation = (False, False, False)
    pbone.lock_rotation_w = False
    pbone.lock_scale = (False, False, False)
    pbone.rotation_mode = 'QUATERNION'
    assign_bone_collections(pbone, 'Leg.L (IK)')
    pbone = obj.pose.bones[bones['toe.R']]
    pbone.rigify_type = ''
    pbone.lock_location = (False, False, False)
    pbone.lock_rotation = (False, False, False)
    pbone.lock_rotation_w = False
    pbone.lock_scale = (False, False, False)
    pbone.rotation_mode = 'QUATERNION'
    assign_bone_collections(pbone, 'Leg.R (IK)')
    pbone = obj.pose.bones[bones['Wing.002.L']]
    pbone.rigify_type = ''
    pbone.lock_location = (False, False, False)
    pbone.lock_rotation = (False, False, False)
    pbone.lock_rotation_w = False
    pbone.lock_scale = (False, False, False)
    pbone.rotation_mode = 'QUATERNION'
    assign_bone_collections(pbone, 'Wing.L')
    pbone = obj.pose.bones[bones['w_feather.003.L']]
    pbone.rigify_type = 'basic.super_copy'
    pbone.lock_location = (False, False, False)
    pbone.lock_rotation = (False, False, False)
    pbone.lock_rotation_w = False
    pbone.lock_scale = (False, False, False)
    pbone.rotation_mode = 'QUATERNION'
    assign_bone_collections(pbone, 'Feathers')
    try:
        pbone.rigify_parameters.make_widget = False
    except AttributeError:
        pass
    pbone = obj.pose.bones[bones['head']]
    pbone.rigify_type = ''
    pbone.lock_location = (False, False, False)
    pbone.lock_rotation = (False, False, False)
    pbone.lock_rotation_w = False
    pbone.lock_scale = (False, False, False)
    pbone.rotation_mode = 'QUATERNION'
    assign_bone_collections(pbone, 'Spine')
    pbone = obj.pose.bones[bones['Wing.002.R']]
    pbone.rigify_type = ''
    pbone.lock_location = (False, False, False)
    pbone.lock_rotation = (False, False, False)
    pbone.lock_rotation_w = False
    pbone.lock_scale = (False, False, False)
    pbone.rotation_mode = 'QUATERNION'
    assign_bone_collections(pbone, 'Wing.R')
    pbone = obj.pose.bones[bones['w_feather.003.R']]
    pbone.rigify_type = 'basic.super_copy'
    pbone.lock_location = (False, False, False)
    pbone.lock_rotation = (False, False, False)
    pbone.lock_rotation_w = False
    pbone.lock_scale = (False, False, False)
    pbone.rotation_mode = 'QUATERNION'
    assign_bone_collections(pbone, 'Feathers')
    try:
        pbone.rigify_parameters.make_widget = False
    except AttributeError:
        pass
    pbone = obj.pose.bones[bones['toes_parent.L']]
    pbone.rigify_type = ''
    pbone.lock_location = (False, False, False)
    pbone.lock_rotation = (False, False, False)
    pbone.lock_rotation_w = False
    pbone.lock_scale = (False, False, False)
    pbone.rotation_mode = 'QUATERNION'
    assign_bone_collections(pbone, 'Claws')
    pbone = obj.pose.bones[bones['t_thumb.001.L']]
    pbone.rigify_type = 'limbs.simple_tentacle'
    pbone.lock_location = (False, False, False)
    pbone.lock_rotation = (False, False, False)
    pbone.lock_rotation_w = False
    pbone.lock_scale = (False, False, False)
    pbone.rotation_mode = 'QUATERNION'
    assign_bone_collections(pbone, 'Claws')
    assign_bone_collection_refs(pbone.rigify_parameters, 'tweak', 'Claws (Tweak)')
    pbone = obj.pose.bones[bones['toes_parent.R']]
    pbone.rigify_type = ''
    pbone.lock_location = (False, False, False)
    pbone.lock_rotation = (False, False, False)
    pbone.lock_rotation_w = False
    pbone.lock_scale = (False, False, False)
    pbone.rotation_mode = 'QUATERNION'
    assign_bone_collections(pbone, 'Claws')
    pbone = obj.pose.bones[bones['t_thumb.001.R']]
    pbone.rigify_type = 'limbs.simple_tentacle'
    pbone.lock_location = (False, False, False)
    pbone.lock_rotation = (False, False, False)
    pbone.lock_rotation_w = False
    pbone.lock_scale = (False, False, False)
    pbone.rotation_mode = 'QUATERNION'
    assign_bone_collections(pbone, 'Claws')
    assign_bone_collection_refs(pbone.rigify_parameters, 'tweak', 'Claws (Tweak)')
    pbone = obj.pose.bones[bones['w_feather.001.L']]
    pbone.rigify_type = 'basic.super_copy'
    pbone.lock_location = (False, False, False)
    pbone.lock_rotation = (False, False, False)
    pbone.lock_rotation_w = False
    pbone.lock_scale = (False, False, False)
    pbone.rotation_mode = 'QUATERNION'
    assign_bone_collections(pbone, 'Feathers')
    try:
        pbone.rigify_parameters.make_widget = False
    except AttributeError:
        pass
    pbone = obj.pose.bones[bones['w_feather.002.L']]
    pbone.rigify_type = 'basic.super_copy'
    pbone.lock_location = (False, False, False)
    pbone.lock_rotation = (False, False, False)
    pbone.lock_rotation_w = False
    pbone.lock_scale = (False, False, False)
    pbone.rotation_mode = 'QUATERNION'
    assign_bone_collections(pbone, 'Feathers')
    try:
        pbone.rigify_parameters.make_widget = False
    except AttributeError:
        pass
    pbone = obj.pose.bones[bones['beak.001.T']]
    pbone.rigify_type = 'limbs.simple_tentacle'
    pbone.lock_location = (False, False, False)
    pbone.lock_rotation = (False, False, False)
    pbone.lock_rotation_w = False
    pbone.lock_scale = (False, False, False)
    pbone.rotation_mode = 'QUATERNION'
    assign_bone_collections(pbone, 'Face')
    assign_bone_collection_refs(pbone.rigify_parameters, 'tweak', 'Face (Tweak)')
    pbone = obj.pose.bones[bones['beak_001.B']]
    pbone.rigify_type = 'limbs.simple_tentacle'
    pbone.lock_location = (False, False, False)
    pbone.lock_rotation = (False, False, False)
    pbone.lock_rotation_w = False
    pbone.lock_scale = (False, False, False)
    pbone.rotation_mode = 'QUATERNION'
    assign_bone_collections(pbone, 'Face')
    assign_bone_collection_refs(pbone.rigify_parameters, 'tweak', 'Face (Tweak)')
    pbone = obj.pose.bones[bones['eye.L']]
    pbone.rigify_type = 'basic.super_copy'
    pbone.lock_location = (False, False, False)
    pbone.lock_rotation = (False, False, False)
    pbone.lock_rotation_w = False
    pbone.lock_scale = (False, False, False)
    pbone.rotation_mode = 'QUATERNION'
    assign_bone_collections(pbone, 'Face')
    try:
        pbone.rigify_parameters.make_widget = False
    except AttributeError:
        pass
    pbone = obj.pose.bones[bones['eye.R']]
    pbone.rigify_type = 'basic.super_copy'
    pbone.lock_location = (False, False, False)
    pbone.lock_rotation = (False, False, False)
    pbone.lock_rotation_w = False
    pbone.lock_scale = (False, False, False)
    pbone.rotation_mode = 'QUATERNION'
    assign_bone_collections(pbone, 'Face')
    try:
        pbone.rigify_parameters.make_widget = False
    except AttributeError:
        pass
    pbone = obj.pose.bones[bones['n_feather.001.L']]
    pbone.rigify_type = 'basic.super_copy'
    pbone.lock_location = (False, False, False)
    pbone.lock_rotation = (False, False, False)
    pbone.lock_rotation_w = False
    pbone.lock_scale = (False, False, False)
    pbone.rotation_mode = 'QUATERNION'
    assign_bone_collections(pbone, 'Feathers')
    try:
        pbone.rigify_parameters.make_widget = False
    except AttributeError:
        pass
    pbone = obj.pose.bones[bones['n_feather.001.R']]
    pbone.rigify_type = 'basic.super_copy'
    pbone.lock_location = (False, False, False)
    pbone.lock_rotation = (False, False, False)
    pbone.lock_rotation_w = False
    pbone.lock_scale = (False, False, False)
    pbone.rotation_mode = 'QUATERNION'
    assign_bone_collections(pbone, 'Feathers')
    try:
        pbone.rigify_parameters.make_widget = False
    except AttributeError:
        pass
    pbone = obj.pose.bones[bones['skull.006.L']]
    pbone.rigify_type = 'basic.super_copy'
    pbone.lock_location = (False, False, False)
    pbone.lock_rotation = (False, False, False)
    pbone.lock_rotation_w = False
    pbone.lock_scale = (False, False, False)
    pbone.rotation_mode = 'QUATERNION'
    assign_bone_collections(pbone, 'Face')
    try:
        pbone.rigify_parameters.make_widget = False
    except AttributeError:
        pass
    try:
        pbone.rigify_parameters.make_control = False
    except AttributeError:
        pass
    pbone = obj.pose.bones[bones['skull.006.R']]
    pbone.rigify_type = 'basic.super_copy'
    pbone.lock_location = (False, False, False)
    pbone.lock_rotation = (False, False, False)
    pbone.lock_rotation_w = False
    pbone.lock_scale = (False, False, False)
    pbone.rotation_mode = 'QUATERNION'
    assign_bone_collections(pbone, 'Face')
    try:
        pbone.rigify_parameters.make_widget = False
    except AttributeError:
        pass
    try:
        pbone.rigify_parameters.make_control = False
    except AttributeError:
        pass
    pbone = obj.pose.bones[bones['w_feather.001.R']]
    pbone.rigify_type = 'basic.super_copy'
    pbone.lock_location = (False, False, False)
    pbone.lock_rotation = (False, False, False)
    pbone.lock_rotation_w = False
    pbone.lock_scale = (False, False, False)
    pbone.rotation_mode = 'QUATERNION'
    assign_bone_collections(pbone, 'Feathers')
    try:
        pbone.rigify_parameters.make_widget = False
    except AttributeError:
        pass
    pbone = obj.pose.bones[bones['w_feather.002.R']]
    pbone.rigify_type = 'basic.super_copy'
    pbone.lock_location = (False, False, False)
    pbone.lock_rotation = (False, False, False)
    pbone.lock_rotation_w = False
    pbone.lock_scale = (False, False, False)
    pbone.rotation_mode = 'QUATERNION'
    assign_bone_collections(pbone, 'Feathers')
    try:
        pbone.rigify_parameters.make_widget = False
    except AttributeError:
        pass
    pbone = obj.pose.bones[bones['t_ring.001.L']]
    pbone.rigify_type = 'limbs.simple_tentacle'
    pbone.lock_location = (False, False, False)
    pbone.lock_rotation = (False, False, False)
    pbone.lock_rotation_w = False
    pbone.lock_scale = (False, False, False)
    pbone.rotation_mode = 'QUATERNION'
    assign_bone_collections(pbone, 'Claws')
    assign_bone_collection_refs(pbone.rigify_parameters, 'tweak', 'Claws (Tweak)')
    pbone = obj.pose.bones[bones['t_index.001.L']]
    pbone.rigify_type = 'limbs.simple_tentacle'
    pbone.lock_location = (False, False, False)
    pbone.lock_rotation = (False, False, False)
    pbone.lock_rotation_w = False
    pbone.lock_scale = (False, False, False)
    pbone.rotation_mode = 'QUATERNION'
    assign_bone_collections(pbone, 'Claws')
    assign_bone_collection_refs(pbone.rigify_parameters, 'tweak', 'Claws (Tweak)')
    pbone = obj.pose.bones[bones['t_middle.001.L']]
    pbone.rigify_type = 'limbs.simple_tentacle'
    pbone.lock_location = (False, False, False)
    pbone.lock_rotation = (False, False, False)
    pbone.lock_rotation_w = False
    pbone.lock_scale = (False, False, False)
    pbone.rotation_mode = 'QUATERNION'
    assign_bone_collections(pbone, 'Claws')
    assign_bone_collection_refs(pbone.rigify_parameters, 'tweak', 'Claws (Tweak)')
    pbone = obj.pose.bones[bones['t_thumb.002.L']]
    pbone.rigify_type = ''
    pbone.lock_location = (False, False, False)
    pbone.lock_rotation = (False, False, False)
    pbone.lock_rotation_w = False
    pbone.lock_scale = (False, False, False)
    pbone.rotation_mode = 'QUATERNION'
    assign_bone_collections(pbone, 'Claws')
    pbone = obj.pose.bones[bones['t_ring.001.R']]
    pbone.rigify_type = 'limbs.simple_tentacle'
    pbone.lock_location = (False, False, False)
    pbone.lock_rotation = (False, False, False)
    pbone.lock_rotation_w = False
    pbone.lock_scale = (False, False, False)
    pbone.rotation_mode = 'QUATERNION'
    assign_bone_collections(pbone, 'Claws')
    assign_bone_collection_refs(pbone.rigify_parameters, 'tweak', 'Claws (Tweak)')
    pbone = obj.pose.bones[bones['t_index.001.R']]
    pbone.rigify_type = 'limbs.simple_tentacle'
    pbone.lock_location = (False, False, False)
    pbone.lock_rotation = (False, False, False)
    pbone.lock_rotation_w = False
    pbone.lock_scale = (False, False, False)
    pbone.rotation_mode = 'QUATERNION'
    assign_bone_collections(pbone, 'Claws')
    assign_bone_collection_refs(pbone.rigify_parameters, 'tweak', 'Claws (Tweak)')
    pbone = obj.pose.bones[bones['t_middle.001.R']]
    pbone.rigify_type = 'limbs.simple_tentacle'
    pbone.lock_location = (False, False, False)
    pbone.lock_rotation = (False, False, False)
    pbone.lock_rotation_w = False
    pbone.lock_scale = (False, False, False)
    pbone.rotation_mode = 'QUATERNION'
    assign_bone_collections(pbone, 'Claws')
    assign_bone_collection_refs(pbone.rigify_parameters, 'tweak', 'Claws (Tweak)')
    pbone = obj.pose.bones[bones['t_thumb.002.R']]
    pbone.rigify_type = ''
    pbone.lock_location = (False, False, False)
    pbone.lock_rotation = (False, False, False)
    pbone.lock_rotation_w = False
    pbone.lock_scale = (False, False, False)
    pbone.rotation_mode = 'QUATERNION'
    assign_bone_collections(pbone, 'Claws')
    pbone = obj.pose.bones[bones['beak.002.T']]
    pbone.rigify_type = ''
    pbone.lock_location = (False, False, False)
    pbone.lock_rotation = (False, False, False)
    pbone.lock_rotation_w = False
    pbone.lock_scale = (False, False, False)
    pbone.rotation_mode = 'QUATERNION'
    assign_bone_collections(pbone, 'Face')
    pbone = obj.pose.bones[bones['beak.002.B']]
    pbone.rigify_type = ''
    pbone.lock_location = (False, False, False)
    pbone.lock_rotation = (False, False, False)
    pbone.lock_rotation_w = False
    pbone.lock_scale = (False, False, False)
    pbone.rotation_mode = 'QUATERNION'
    assign_bone_collections(pbone, 'Face')
    pbone = obj.pose.bones[bones['tongue.001.L']]
    pbone.rigify_type = 'limbs.simple_tentacle'
    pbone.lock_location = (False, False, False)
    pbone.lock_rotation = (False, False, False)
    pbone.lock_rotation_w = False
    pbone.lock_scale = (False, False, False)
    pbone.rotation_mode = 'QUATERNION'
    assign_bone_collections(pbone, 'Face')
    assign_bone_collection_refs(pbone.rigify_parameters, 'tweak', 'Face (Tweak)')
    pbone = obj.pose.bones[bones['t_ring.002.L']]
    pbone.rigify_type = ''
    pbone.lock_location = (False, False, False)
    pbone.lock_rotation = (False, False, False)
    pbone.lock_rotation_w = False
    pbone.lock_scale = (False, False, False)
    pbone.rotation_mode = 'QUATERNION'
    assign_bone_collections(pbone, 'Claws')
    pbone = obj.pose.bones[bones['t_index.002.L']]
    pbone.rigify_type = ''
    pbone.lock_location = (False, False, False)
    pbone.lock_rotation = (False, False, False)
    pbone.lock_rotation_w = False
    pbone.lock_scale = (False, False, False)
    pbone.rotation_mode = 'QUATERNION'
    assign_bone_collections(pbone, 'Claws')
    pbone = obj.pose.bones[bones['t_middle.002.L']]
    pbone.rigify_type = ''
    pbone.lock_location = (False, False, False)
    pbone.lock_rotation = (False, False, False)
    pbone.lock_rotation_w = False
    pbone.lock_scale = (False, False, False)
    pbone.rotation_mode = 'QUATERNION'
    assign_bone_collections(pbone, 'Claws')
    pbone = obj.pose.bones[bones['t_ring.002.R']]
    pbone.rigify_type = ''
    pbone.lock_location = (False, False, False)
    pbone.lock_rotation = (False, False, False)
    pbone.lock_rotation_w = False
    pbone.lock_scale = (False, False, False)
    pbone.rotation_mode = 'QUATERNION'
    assign_bone_collections(pbone, 'Claws')
    pbone = obj.pose.bones[bones['t_index.002.R']]
    pbone.rigify_type = ''
    pbone.lock_location = (False, False, False)
    pbone.lock_rotation = (False, False, False)
    pbone.lock_rotation_w = False
    pbone.lock_scale = (False, False, False)
    pbone.rotation_mode = 'QUATERNION'
    assign_bone_collections(pbone, 'Claws')
    pbone = obj.pose.bones[bones['t_middle.002.R']]
    pbone.rigify_type = ''
    pbone.lock_location = (False, False, False)
    pbone.lock_rotation = (False, False, False)
    pbone.lock_rotation_w = False
    pbone.lock_scale = (False, False, False)
    pbone.rotation_mode = 'QUATERNION'
    assign_bone_collections(pbone, 'Claws')
    pbone = obj.pose.bones[bones['tongue.002.L']]
    pbone.rigify_type = ''
    pbone.lock_location = (False, False, False)
    pbone.lock_rotation = (False, False, False)
    pbone.lock_rotation_w = False
    pbone.lock_scale = (False, False, False)
    pbone.rotation_mode = 'QUATERNION'
    assign_bone_collections(pbone, 'Face')
    pbone = obj.pose.bones[bones['t_ring.003.L']]
    pbone.rigify_type = ''
    pbone.lock_location = (False, False, False)
    pbone.lock_rotation = (False, False, False)
    pbone.lock_rotation_w = False
    pbone.lock_scale = (False, False, False)
    pbone.rotation_mode = 'QUATERNION'
    assign_bone_collections(pbone, 'Claws')
    pbone = obj.pose.bones[bones['t_index.003.L']]
    pbone.rigify_type = ''
    pbone.lock_location = (False, False, False)
    pbone.lock_rotation = (False, False, False)
    pbone.lock_rotation_w = False
    pbone.lock_scale = (False, False, False)
    pbone.rotation_mode = 'QUATERNION'
    assign_bone_collections(pbone, 'Claws')
    pbone = obj.pose.bones[bones['t_middle.003.L']]
    pbone.rigify_type = ''
    pbone.lock_location = (False, False, False)
    pbone.lock_rotation = (False, False, False)
    pbone.lock_rotation_w = False
    pbone.lock_scale = (False, False, False)
    pbone.rotation_mode = 'QUATERNION'
    assign_bone_collections(pbone, 'Claws')
    pbone = obj.pose.bones[bones['t_ring.003.R']]
    pbone.rigify_type = ''
    pbone.lock_location = (False, False, False)
    pbone.lock_rotation = (False, False, False)
    pbone.lock_rotation_w = False
    pbone.lock_scale = (False, False, False)
    pbone.rotation_mode = 'QUATERNION'
    assign_bone_collections(pbone, 'Claws')
    pbone = obj.pose.bones[bones['t_index.003.R']]
    pbone.rigify_type = ''
    pbone.lock_location = (False, False, False)
    pbone.lock_rotation = (False, False, False)
    pbone.lock_rotation_w = False
    pbone.lock_scale = (False, False, False)
    pbone.rotation_mode = 'QUATERNION'
    assign_bone_collections(pbone, 'Claws')
    pbone = obj.pose.bones[bones['t_middle.003.R']]
    pbone.rigify_type = ''
    pbone.lock_location = (False, False, False)
    pbone.lock_rotation = (False, False, False)
    pbone.lock_rotation_w = False
    pbone.lock_scale = (False, False, False)
    pbone.rotation_mode = 'QUATERNION'
    assign_bone_collections(pbone, 'Claws')
    pbone = obj.pose.bones[bones['tongue.003.L']]
    pbone.rigify_type = ''
    pbone.lock_location = (False, False, False)
    pbone.lock_rotation = (False, False, False)
    pbone.lock_rotation_w = False
    pbone.lock_scale = (False, False, False)
    pbone.rotation_mode = 'QUATERNION'
    assign_bone_collections(pbone, 'Face')

    bpy.ops.object.mode_set(mode='EDIT')
    for bone in arm.edit_bones:
        bone.select = False
        bone.select_head = False
        bone.select_tail = False
    for b in bones:
        bone = arm.edit_bones[bones[b]]
        bone.select = True
        bone.select_head = True
        bone.select_tail = True
        bone.bbone_x = bone.bbone_z = bone.length * 0.05
        arm.edit_bones.active = bone

    arm.collections.active_index = 0

    return bones


if __name__ == "__main__":
    create(bpy.context.active_object)



# SPDX-FileCopyrightText: 2018-2021 The glTF-Blender-IO authors
#
# SPDX-License-Identifier: Apache-2.0

from ....blender.com.data_path import get_sk_exported, skip_sk, get_target_object_path
from ..cache import skdriverdiscovercache


@skdriverdiscovercache
def get_sk_drivers(blender_armature_uuid, export_settings):

    # If no SK are exported --> No driver animation to export
    if export_settings['gltf_morph_anim'] is False:
        return []

    drivers = []

    # Take into account skinned mesh, and mesh parented to a bone of the armature
    children_list = export_settings['vtree'].nodes[blender_armature_uuid].children.copy()
    for bone in export_settings['vtree'].get_all_bones(blender_armature_uuid):
        children_list.extend(export_settings['vtree'].nodes[bone].children)

    for child_uuid in children_list:
        uuid, channels = get_driver_on_shapekey(child_uuid, export_settings)
        if uuid is not None:
            drivers.append(child_uuid)

    return drivers


def get_driver_on_shapekey(blender_object_uuid, export_settings):
    if export_settings['vtree'].nodes[blender_object_uuid].blender_type == "BONE":
        return None, None

    child = export_settings['vtree'].nodes[blender_object_uuid].blender_object

    if not child.data:
        return None, None
    # child.data can be an armature - which has no shapekeys
    if not hasattr(child.data, 'shape_keys'):
        return None, None
    if not child.data.shape_keys:
        return None, None
    if not child.data.shape_keys.animation_data:
        return None, None
    if not child.data.shape_keys.animation_data.drivers:
        return None, None
    if len(child.data.shape_keys.animation_data.drivers) <= 0:
        return None, None

    shapekeys_idx = {}
    cpt_sk = 0
    for sk in get_sk_exported(child.data.shape_keys.key_blocks):
        shapekeys_idx[sk.name] = cpt_sk
        cpt_sk += 1

    # Note: channels will have some None items only for SK if some SK are not animated
    idx_channel_mapping = []
    all_sorted_channels = []
    for sk_c in child.data.shape_keys.animation_data.drivers:
        # Check if driver is valid. If not, ignore this driver channel
        try:
            # Check if driver is valid.
            # Try/Except is no more a suffisant check, starting with version Blender 3.0,
            # Blender crashes when trying to resolve path on invalid driver
            if not sk_c.is_valid:
                return None, None
            sk_name = child.data.shape_keys.path_resolve(get_target_object_path(sk_c.data_path)).name
        except:
            return None, None
        if skip_sk(child.data.shape_keys.key_blocks, child.data.shape_keys.key_blocks[sk_name]):
            return None, None
        idx_channel_mapping.append((shapekeys_idx[sk_name], sk_c))
    existing_idx = dict(idx_channel_mapping)
    for i in range(0, cpt_sk):
        if i not in existing_idx.keys():
            all_sorted_channels.append(None)
        else:
            all_sorted_channels.append(existing_idx[i])

    # Checks there are some driver on SK, and that there is not only invalid drivers
    if len(all_sorted_channels) > 0 and not all([i is None for i in all_sorted_channels]):
        return blender_object_uuid, all_sorted_channels



# SPDX-FileCopyrightText: 2018-2022 The glTF-Blender-IO authors
#
# SPDX-License-Identifier: Apache-2.0

import mathutils
import bpy
import typing
from .....blender.com.data_path import get_sk_exported
from .....blender.com.conversion import inverted_trs_mapping_node, texture_transform_blender_to_gltf, yvof_blender_to_gltf
from ...cache import datacache
from ...tree import VExportNode
from ..drivers import get_sk_drivers

# Warning : If you change some parameter here, need to be changed in cache system


@datacache
def get_cache_data(path: str,
                   blender_obj_uuid: str,
                   bone: typing.Optional[str],
                   action_name: str,
                   current_frame: int,
                   step: int,
                   slot_identifier: str,
                   export_settings,
                   only_gather_provided=False
                   ):

    data = {}

    # Ranges are stored at action level, so no need to give the slot_identifier here
    min_, max_ = get_range(blender_obj_uuid, action_name, export_settings)

    if only_gather_provided:
        # If object is not in vtree, this is a material or light for pointers
        obj_uuids = [blender_obj_uuid] if blender_obj_uuid in export_settings['vtree'].nodes.keys() else []
    else:
        obj_uuids = [uid for (uid, n) in export_settings['vtree'].nodes.items()
                     if n.blender_type not in [VExportNode.BONE]]

    # For TRACK mode, we reset cache after each track export, so we don't need to keep others objects
    if export_settings['gltf_animation_mode'] in ["NLA_TRACKS"]:
        # If object is not in vtree, this is a material or light for pointers
        obj_uuids = [blender_obj_uuid] if blender_obj_uuid in export_settings['vtree'].nodes.keys() else []

    # If there is only 1 object to cache, we can disable viewport for other objects (for performance)
    # This can be on these cases:
    # - TRACK mode
    # - Only one object to cache (but here, no really useful for performance)
    # - Action mode, where some object have multiple actions
        # - For this case, on first call, we will cache active action for all objects
        # - On next calls, we will cache only the action of current object, so we can disable viewport for others

    need_to_enable_again = False
    if export_settings['gltf_optimize_disable_viewport'] is True and len(obj_uuids) == 1:
        need_to_enable_again = True
        # Before baking, disabling from viewport all meshes
        for obj in [n.blender_object for n in export_settings['vtree'].nodes.values() if n.blender_type in
                    [VExportNode.OBJECT, VExportNode.ARMATURE, VExportNode.COLLECTION]]:
            if obj is None:
                continue
            obj.hide_viewport = True
        export_settings['vtree'].nodes[obj_uuids[0]].blender_object.hide_viewport = False
    # Work for drivers on shapekeys is already done at start of animation export

    depsgraph = bpy.context.evaluated_depsgraph_get()

    frame = min_
    while frame <= max_:
        bpy.context.scene.frame_set(int(frame))
        current_instance = {}  # For GN instances, we are going to track instances by their order in instance iterator

        object_caching(data, obj_uuids, current_instance, action_name, slot_identifier, frame, depsgraph, export_settings)

        # KHR_animation_pointer caching for materials, lights, cameras
        if export_settings['gltf_export_anim_pointer'] is True:
            material_nodetree_caching(data, action_name, slot_identifier, frame, export_settings)
            material_caching(data, action_name, slot_identifier, frame, export_settings)
            light_nodetree_caching(data, action_name, slot_identifier, frame, export_settings)
            camera_caching(data, action_name, slot_identifier, frame, export_settings)

        frame += step

    # And now, restoring meshes in viewport
    for node, obj in [(n, n.blender_object) for n in export_settings['vtree'].nodes.values() if n.blender_type in
                      [VExportNode.OBJECT, VExportNode.ARMATURE, VExportNode.COLLECTION]]:
        obj.hide_viewport = node.default_hide_viewport

    return data

# For perf, we may be more precise, and get a list of ranges to be exported that include all needed frames


def get_range(obj_uuid, key, export_settings):
    if export_settings['gltf_animation_mode'] in ["NLA_TRACKS"]:
        return export_settings['ranges'][obj_uuid][key]['start'], export_settings['ranges'][obj_uuid][key]['end']
    else:
        min_ = None
        max_ = None
        for obj in export_settings['ranges'].keys():
            for anim in export_settings['ranges'][obj].keys():
                if min_ is None or min_ > export_settings['ranges'][obj][anim]['start']:
                    min_ = export_settings['ranges'][obj][anim]['start']
                if max_ is None or max_ < export_settings['ranges'][obj][anim]['end']:
                    max_ = export_settings['ranges'][obj][anim]['end']
    return min_, max_


def initialize_data_dict(data, key1, key2, key3, key4, key5):
    # No check on key1, this is already done before calling this function
    if key2 not in data[key1].keys():
        data[key1][key2] = {}
        data[key1][key2][key3] = {}
        data[key1][key2][key3][key4] = {}
        data[key1][key2][key3][key4][key5] = {}

    if key3 not in data[key1][key2].keys():
        data[key1][key2][key3] = {}
        data[key1][key2][key3][key4] = {}
        data[key1][key2][key3][key4][key5] = {}


def material_caching(data, action_name, slot_identifier, frame, export_settings):
    for mat in export_settings['KHR_animation_pointer']['materials'].keys():
        if len(export_settings['KHR_animation_pointer']['materials'][mat]['paths']) == 0:
            continue

        if export_settings['gltf_animation_mode'] == "NLA_TRACKS" and export_settings['gltf_apply'] is True:
            blender_material = [export_settings['material_identifiers'][mat]]
        else:
            blender_material = [m for m in bpy.data.materials if id(m) == mat]
        if len(blender_material) == 0:
            # This is not a material from Blender (coming from Geometry Node for example, so no animation on it)
            continue
        else:
            blender_material = blender_material[0]
        if mat not in data.keys():
            data[mat] = {}

        if blender_material and blender_material.animation_data and blender_material.animation_data.action \
                and blender_material.animation_data.action_slot \
                and export_settings['gltf_animation_mode'] in ["ACTIVE_ACTIONS", "ACTIONS"]:
            key1, key2, key3, key4 = mat, blender_material.animation_data.action.name, blender_material.animation_data.action_slot.identifier, "value"
        elif export_settings['gltf_animation_mode'] in ["NLA_TRACKS"]:
            # We can keep the input slot_identifier here, as we are caching only one object / NLA track
            key1, key2, key3, key4 = mat, action_name, slot_identifier, "value"
        else:
            # case of baking materials (scene export).
            # There is no animation, so use id as key
            # slot_identifier is always None for scene export
            key1, key2, key3, key4 = mat, mat, slot_identifier, "value"

        if key2 not in data[key1].keys():
            data[key1][key2] = {}
            data[key1][key2][key3] = {}
            data[key1][key2][key3][key4] = {}

            for path in export_settings['KHR_animation_pointer']['materials'][mat]['paths'].keys():
                data[key1][key2][key3][key4][path] = {}

        if key3 not in data[key1][key2].keys():
            data[key1][key2][key3] = {}
            data[key1][key2][key3][key4] = {}

            for path in export_settings['KHR_animation_pointer']['materials'][mat]['paths'].keys():
                data[key1][key2][key3][key4][path] = {}

        for path in export_settings['KHR_animation_pointer']['materials'][mat]['paths'].keys():

            if path.startswith("node_tree"):
                continue

            val = blender_material.path_resolve(path)
            if type(val).__name__ == "float":
                data[key1][key2][key3][key4][path][frame] = val
            else:
                data[key1][key2][key3][key4][path][frame] = list(val)


def material_nodetree_caching(data, action_name, slot_identifier, frame, export_settings):
    # After caching objects, caching materials, for KHR_animation_pointer
    for mat in export_settings['KHR_animation_pointer']['materials'].keys():
        if len(export_settings['KHR_animation_pointer']['materials'][mat]['paths']) == 0:
            continue

        if export_settings['gltf_animation_mode'] == "NLA_TRACKS" and export_settings['gltf_apply'] is True:
            blender_material = [export_settings['material_identifiers'][mat]]
        else:
            blender_material = [m for m in bpy.data.materials if id(m) == mat]
        if len(blender_material) == 0:
            # This is not a material from Blender (coming from Geometry Node for example, so no animation on it)
            continue
        else:
            blender_material = blender_material[0]
        if mat not in data.keys():
            data[mat] = {}

        if blender_material.node_tree and blender_material.node_tree.animation_data and blender_material.node_tree.animation_data.action \
                and blender_material.node_tree.animation_data.action_slot \
                and export_settings['gltf_animation_mode'] in ["ACTIVE_ACTIONS", "ACTIONS"]:
            key1, key2, key3, key4 = mat, blender_material.node_tree.animation_data.action.name, blender_material.node_tree.animation_data.action_slot.identifier, "value"
        elif export_settings['gltf_animation_mode'] in ["NLA_TRACKS"]:
            # We can keep the input slot_identifier here, as we are caching only one object / NLA track
            key1, key2, key3, key4 = mat, action_name, slot_identifier, "value"
        else:
            # case of baking materials (scene export).
            # There is no animation, so use id as key
            # slot_identifier is always None for scene export
            key1, key2, key3, key4 = mat, mat, slot_identifier, "value"

        if key2 not in data[key1].keys():
            data[key1][key2] = {}
            data[key1][key2][key3] = {}
            data[key1][key2][key3][key4] = {}
            for path in export_settings['KHR_animation_pointer']['materials'][mat]['paths'].keys():
                data[key1][key2][key3][key4][path] = {}

        if key3 not in data[key1][key2].keys():
            data[key1][key2][key3] = {}
            data[key1][key2][key3][key4] = {}
            for path in export_settings['KHR_animation_pointer']['materials'][mat]['paths'].keys():
                data[key1][key2][key3][key4][path] = {}

        baseColorFactor_alpha_merged_already_done = False
        for path in export_settings['KHR_animation_pointer']['materials'][mat]['paths'].keys():

            if not path.startswith("node_tree"):
                continue

            # Manage special case where we merge baseColorFactor and alpha
            if export_settings['KHR_animation_pointer']['materials'][mat]['paths'][path]['path'] == "/materials/XXX/pbrMetallicRoughness/baseColorFactor" \
                    and export_settings['KHR_animation_pointer']['materials'][mat]['paths'][path]['length'] == 3:
                if baseColorFactor_alpha_merged_already_done is True:
                    continue
                val_color = blender_material.path_resolve(path)
                data_color = list(val_color)[:export_settings['KHR_animation_pointer']
                                             ['materials'][mat]['paths'][path]['length']]
                if export_settings['KHR_animation_pointer']['materials'][mat]['paths'][path]['additional_path'] is not None:
                    val_alpha = blender_material.path_resolve(
                        export_settings['KHR_animation_pointer']['materials'][mat]['paths'][path]['additional_path'])
                else:
                    val_alpha = 1.0
                data[key1][key2][key3][key4][path][frame] = data_color + [val_alpha]
                baseColorFactor_alpha_merged_already_done = True
            # Manage special case where we merge baseColorFactor and alpha
            elif export_settings['KHR_animation_pointer']['materials'][mat]['paths'][path]['path'] == "/materials/XXX/pbrMetallicRoughness/baseColorFactor" \
                    and export_settings['KHR_animation_pointer']['materials'][mat]['paths'][path]['length'] == 1:
                if baseColorFactor_alpha_merged_already_done is True:
                    continue
                val_alpha = blender_material.path_resolve(path)
                if export_settings['KHR_animation_pointer']['materials'][mat]['paths'][path]['additional_path'] is not None:
                    val_color = blender_material.path_resolve(
                        export_settings['KHR_animation_pointer']['materials'][mat]['paths'][path]['additional_path'])
                    data_color = list(val_color)[:export_settings['KHR_animation_pointer']
                                                 ['materials'][mat]['paths']['additional_path']['length']]
                else:
                    data_color = [1.0, 1.0, 1.0]
                data[key1][key2][key3][key4][path][frame] = data_color + [val_alpha]
                baseColorFactor_alpha_merged_already_done = True

            # Manage special case for KHR_texture_transform offset, that needs
            # rotation and scale too (and not only translation)
            elif "KHR_texture_transform" in export_settings['KHR_animation_pointer']['materials'][mat]['paths'][path]['path'] \
                    and export_settings['KHR_animation_pointer']['materials'][mat]['paths'][path]['path'].endswith("offset"):

                val_offset = blender_material.path_resolve(path)
                rotation_path = [
                    i for i in export_settings['KHR_animation_pointer']['materials'][mat]['paths'].keys() if export_settings['KHR_animation_pointer']['materials'][mat]['paths'][i]['path'].rsplit(
                        "/",
                        1)[0] == export_settings['KHR_animation_pointer']['materials'][mat]['paths'][path]['path'].rsplit(
                        "/",
                        1)[0] and export_settings['KHR_animation_pointer']['materials'][mat]['paths'][i]['path'].rsplit(
                        "/",
                        1)[1] == "rotation"][0]
                val_rotation = blender_material.path_resolve(rotation_path)
                scale_path = [
                    i for i in export_settings['KHR_animation_pointer']['materials'][mat]['paths'].keys() if export_settings['KHR_animation_pointer']['materials'][mat]['paths'][i]['path'].rsplit(
                        "/",
                        1)[0] == export_settings['KHR_animation_pointer']['materials'][mat]['paths'][path]['path'].rsplit(
                        "/",
                        1)[0] and export_settings['KHR_animation_pointer']['materials'][mat]['paths'][i]['path'].rsplit(
                        "/",
                        1)[1] == "scale"][0]
                val_scale = blender_material.path_resolve(scale_path)

                mapping_transform = {}
                mapping_transform["offset"] = [val_offset[0], val_offset[1]]
                mapping_transform["rotation"] = val_rotation
                mapping_transform["scale"] = [val_scale[0], val_scale[1]]

                if export_settings['KHR_animation_pointer']['materials'][mat]['paths'][path]['vector_type'] == "TEXTURE":
                    mapping_transform = inverted_trs_mapping_node(mapping_transform)
                    if mapping_transform is None:
                        # Can not be converted to TRS, so ... keeping default values
                        export_settings['log'].warning(
                            "Can not convert texture transform to TRS. Keeping default values.")
                        mapping_transform = {}
                        mapping_transform["offset"] = [0.0, 0.0]
                        mapping_transform["rotation"] = 0.0
                        mapping_transform["scale"] = [1.0, 1.0]
                elif export_settings['KHR_animation_pointer']['materials'][mat]['paths'][path]['vector_type'] == "VECTOR":
                    # Vectors don't get translated
                    mapping_transform["offset"] = [0, 0]

                texture_transform = texture_transform_blender_to_gltf(mapping_transform)

                data[key1][key2][key3][key4][path][frame] = texture_transform['offset']
                data[key1][key2][key3][key4][rotation_path][frame] = texture_transform['rotation']
                data[key1][key2][key3][key4][scale_path][frame] = texture_transform['scale']
                if export_settings['KHR_animation_pointer']['materials'][mat]['paths'][path]['vector_type'] != "VECTOR":
                    # Already handled by offset
                    continue
                else:
                    val = blender_material.path_resolve(path)
                    mapping_transform = {}
                    mapping_transform["offset"] = [0, 0]  # Placeholder, not needed
                    mapping_transform["rotation"] = val
                    mapping_transform["scale"] = [1, 1]  # Placeholder, not needed
                    texture_transform = texture_transform_blender_to_gltf(mapping_transform)
                    data[key1][key2][key3][key4][path][frame] = texture_transform['rotation']
            elif "KHR_texture_transform" in export_settings['KHR_animation_pointer']['materials'][mat]['paths'][path]['path'] \
                    and export_settings['KHR_animation_pointer']['materials'][mat]['paths'][path]['path'].endswith("scale"):
                if export_settings['KHR_animation_pointer']['materials'][mat]['paths'][path]['vector_type'] != "VECTOR":
                    # Already handled by offset
                    continue
                else:
                    val = blender_material.path_resolve(path)
                    mapping_transform = {}
                    mapping_transform["offset"] = [0, 0]  # Placeholder, not needed
                    mapping_transform["rotation"] = 0.0  # Placeholder, not needed
                    mapping_transform["scale"] = [val[0], val[1]]
                    texture_transform = texture_transform_blender_to_gltf(mapping_transform)
                    data[key1][key2][key3][key4][path][frame] = texture_transform['rotation']

            # Manage special cases for specularFactor & specularColorFactor
            elif export_settings['KHR_animation_pointer']['materials'][mat]['paths'][path]['path'] == "/materials/XXX/extensions/KHR_materials_specular/specularFactor":
                val = blender_material.path_resolve(path)
                val = val * 2.0
                if val > 1.0:
                    fac = val
                    val = 1.0
                else:
                    fac = 1.0

                data[key1][key2][key3][key4][path][frame] = val

                # Retrieve specularColorFactor
                colorfactor_path = [
                    i for i in export_settings['KHR_animation_pointer']['materials'][mat]['paths'].keys() if export_settings['KHR_animation_pointer']['materials'][mat]['paths'][i]['path'].rsplit(
                        "/",
                        1)[0] == export_settings['KHR_animation_pointer']['materials'][mat]['paths'][path]['path'].rsplit(
                        "/",
                        1)[0] and export_settings['KHR_animation_pointer']['materials'][mat]['paths'][i]['path'].rsplit(
                        "/",
                        1)[1] == "specularColorFactor"]
                if len(colorfactor_path) == 0:
                    # No specularColorFactor, so using default value
                    val_colorfactor = [1.0, 1.0, 1.0]
                else:
                    val_colorfactor = blender_material.path_resolve(colorfactor_path[0])
                    colorfactor_path = colorfactor_path[0]
                if fac > 1.0:
                    val_colorfactor = [i * fac for i in val_colorfactor]
                    if len(colorfactor_path) == 0:
                        export_settings['log'].warning(
                            "SpecularFactor is greater than 1.0, but no specularColorFactor found. Using default value for specularColorFactor.")
                if len(colorfactor_path) != 0:
                    data[key1][key2][key3][key4][colorfactor_path][frame] = val_colorfactor
            elif export_settings['KHR_animation_pointer']['materials'][mat]['paths'][path]['path'] == "/materials/XXX/extensions/KHR_materials_specular/specularColorFactor":
                # Already handled by specularFactor
                continue

            # Classic case
            else:
                val = blender_material.path_resolve(path)
                if type(val).__name__ == "float":
                    data[key1][key2][key3][key4][path][frame] = val
                else:
                    data[key1][key2][key3][key4][path][frame] = list(val)[
                        :export_settings['KHR_animation_pointer']['materials'][mat]['paths'][path]['length']]


def armature_caching(data, obj_uuid, blender_obj, action_name, slot_identifier, frame, export_settings):
    bones = export_settings['vtree'].get_all_bones(obj_uuid)
    if blender_obj.animation_data and blender_obj.animation_data.action \
            and blender_obj.animation_data.action_slot \
            and export_settings['gltf_animation_mode'] in ["ACTIVE_ACTIONS", "ACTIONS", "BROADCAST"]:
        key1, key2, key3, key4 = obj_uuid, blender_obj.animation_data.action.name, blender_obj.animation_data.action_slot.identifier, "bone"
    elif blender_obj.animation_data \
            and export_settings['gltf_animation_mode'] in ["NLA_TRACKS"]:
        # We can keep the input slot_identifier here, as we are caching only one object / NLA track
        key1, key2, key3, key4 = obj_uuid, action_name, slot_identifier, "bone"
    else:
        # slot_identifier is always None for scene export
        key1, key2, key3, key4 = obj_uuid, obj_uuid, slot_identifier, "bone"

    if key3 not in data[key1][key2].keys():
        data[key1][key2][key3] = {}
        data[key1][key2][key3][key4] = {}
    if key4 not in data[key1][key2][key3].keys():
        data[key1][key2][key3][key4] = {}

    for bone_uuid in [bone for bone in bones if export_settings['vtree'].nodes[bone].leaf_reference is None]:
        blender_bone = export_settings['vtree'].nodes[bone_uuid].blender_bone

        if export_settings['vtree'].nodes[bone_uuid].parent_uuid is not None and export_settings['vtree'].nodes[
                export_settings['vtree'].nodes[bone_uuid].parent_uuid].blender_type == VExportNode.BONE:
            blender_bone_parent = export_settings['vtree'].nodes[export_settings['vtree']
                                                                 .nodes[bone_uuid].parent_uuid].blender_bone
            rest_mat = blender_bone_parent.bone.matrix_local.inverted_safe() @ blender_bone.bone.matrix_local
            matrix = rest_mat.inverted_safe() @ blender_bone_parent.matrix.inverted_safe() @ blender_bone.matrix
        else:
            if blender_bone.parent is None:
                matrix = blender_bone.bone.matrix_local.inverted_safe() @ blender_bone.matrix
            else:
                # Bone has a parent, but in export, after filter, is at root of armature
                matrix = blender_bone.matrix.copy()

            # Because there is no armature object, we need to apply the TRS of armature to the root bone
            if export_settings['gltf_armature_object_remove'] is True:
                matrix = matrix @ blender_obj.matrix_world

        if blender_bone.name not in data[key1][key2][key3][key4].keys():
            data[key1][key2][key3][key4][blender_bone.name] = {}
        data[key1][key2][key3][key4][blender_bone.name][frame] = matrix


def object_caching(data, obj_uuids, current_instance, action_name, slot_identifier, frame, depsgraph, export_settings):
    for obj_uuid in obj_uuids:

        # Do not cache real collection
        if export_settings['vtree'].nodes[obj_uuid].blender_type == VExportNode.COLLECTION:
            continue

        blender_obj = export_settings['vtree'].nodes[obj_uuid].blender_object
        if blender_obj is None:  # GN instance
            if export_settings['vtree'].nodes[obj_uuid].parent_uuid not in current_instance.keys():
                current_instance[export_settings['vtree'].nodes[obj_uuid].parent_uuid] = 0

        # TODO: we may want to avoid looping on all objects, but an accurate filter must be found

        # calculate local matrix
        if export_settings['vtree'].nodes[obj_uuid].parent_uuid is None:
            parent_mat = mathutils.Matrix.Identity(4).freeze()
        else:
            if export_settings['vtree'].nodes[export_settings['vtree'].nodes[obj_uuid].parent_uuid].blender_type not in [
                    VExportNode.BONE]:
                if export_settings['vtree'].nodes[export_settings['vtree']
                                                  .nodes[obj_uuid].parent_uuid].blender_type != VExportNode.COLLECTION:
                    parent_mat = export_settings['vtree'].nodes[export_settings['vtree']
                                                                .nodes[obj_uuid].parent_uuid].blender_object.matrix_world
                else:
                    parent_mat = export_settings['vtree'].nodes[export_settings['vtree']
                                                                .nodes[obj_uuid].parent_uuid].matrix_world
            else:
                # Object animated is parented to a bone
                blender_bone = export_settings['vtree'].nodes[export_settings['vtree']
                                                              .nodes[obj_uuid].parent_bone_uuid].blender_bone
                armature_object = export_settings['vtree'].nodes[export_settings['vtree']
                                                                 .nodes[export_settings['vtree'].nodes[obj_uuid].parent_bone_uuid].armature].blender_object
                axis_basis_change = mathutils.Matrix(
                    ((1.0, 0.0, 0.0, 0.0), (0.0, 0.0, 1.0, 0.0), (0.0, -1.0, 0.0, 0.0), (0.0, 0.0, 0.0, 1.0)))

                parent_mat = armature_object.matrix_world @ blender_bone.matrix @ axis_basis_change

        # For object inside collection (at root), matrix world is already expressed regarding collection parent
        if export_settings['vtree'].nodes[obj_uuid].parent_uuid is not None and export_settings['vtree'].nodes[
                export_settings['vtree'].nodes[obj_uuid].parent_uuid].blender_type == VExportNode.INST_COLLECTION:
            parent_mat = mathutils.Matrix.Identity(4).freeze()

        if blender_obj:
            mat = parent_mat.inverted_safe() @ blender_obj.matrix_world
        else:
            eval = export_settings['vtree'].nodes[export_settings['vtree'].nodes[obj_uuid].parent_uuid].blender_object.evaluated_get(
                depsgraph)
            cpt_inst = 0
            for inst in depsgraph.object_instances:  # use only as iterator
                if inst.parent == eval:
                    if current_instance[export_settings['vtree'].nodes[obj_uuid].parent_uuid] == cpt_inst:
                        mat = inst.matrix_world.copy()
                        current_instance[export_settings['vtree'].nodes[obj_uuid].parent_uuid] += 1
                        break
                    cpt_inst += 1

        if obj_uuid not in data.keys():
            data[obj_uuid] = {}

        if blender_obj and blender_obj.animation_data and blender_obj.animation_data.action \
                and blender_obj.animation_data.action_slot \
                and export_settings['gltf_animation_mode'] in ["ACTIVE_ACTIONS", "ACTIONS", "BROADCAST"]:
            key1, key2, key3, key4, key5 = obj_uuid, blender_obj.animation_data.action.name, blender_obj.animation_data.action_slot.identifier, "matrix", None
        elif export_settings['gltf_animation_mode'] in ["NLA_TRACKS"]:
            # We can keep the input slot_identifier here, as we are caching only one object / NLA track
            key1, key2, key3, key4, key5 = obj_uuid, action_name, slot_identifier, "matrix", None
        else:
            # case of baking object.
            # There is no animation, so use uuid of object as key
            # slot_identifier is always None for scene export
            key1, key2, key3, key4, key5 = obj_uuid, obj_uuid, slot_identifier, "matrix", None

        initialize_data_dict(data, key1, key2, key3, key4, key5)
        data[key1][key2][key3][key4][key5][frame] = mat

        # Store data for all bones, if object is an armature

        if blender_obj and blender_obj.type == "ARMATURE":
            armature_caching(data, obj_uuid, blender_obj, action_name, slot_identifier, frame, export_settings)

        elif blender_obj is None:  # GN instances
            # case of baking object, for GN instances
            # There is no animation, so use uuid of object as key
            # slot_identifier is always None for baking
            key1, key2, key3, key4, key5 = obj_uuid, obj_uuid, slot_identifier, "matrix", None
            initialize_data_dict(data, key1, key2, key3, key4, key5)
            data[key1][key2][key3][key4][key5][frame] = mat

        # Check SK animation here, as we are caching data
        # This will avoid to have to do it again when exporting SK animation
        cache_sk = False
        if export_settings['gltf_morph_anim'] and blender_obj and blender_obj.type == "MESH" \
                and blender_obj.data is not None \
                and blender_obj.data.shape_keys is not None \
                and blender_obj.data.shape_keys.animation_data is not None \
                and blender_obj.data.shape_keys.animation_data.action is not None \
                and blender_obj.data.shape_keys.animation_data.action_slot is not None \
                and export_settings['gltf_animation_mode'] in ["ACTIVE_ACTIONS", "ACTIONS", "BROADCAST"]:

            key1, key2, key3, key4, key5 = obj_uuid, blender_obj.data.shape_keys.animation_data.action.name, blender_obj.data.shape_keys.animation_data.action_slot.identifier, "sk", None
            cache_sk = True

        elif export_settings['gltf_morph_anim'] and blender_obj and blender_obj.type == "MESH" \
                and blender_obj.data is not None \
                and blender_obj.data.shape_keys is not None \
                and blender_obj.data.shape_keys.animation_data is not None \
                and export_settings['gltf_animation_mode'] in ["NLA_TRACKS"]:

            # We can keep the input slot_identifier here, as we are caching only one object / NLA track
            key1, key2, key3, key4, key5 = obj_uuid, action_name, slot_identifier, "sk", None
            cache_sk = True

        elif export_settings['gltf_morph_anim'] and blender_obj and blender_obj.type == "MESH" \
                and blender_obj.data is not None \
                and blender_obj.data.shape_keys is not None:
            # slot_identifier is always None for scene export
            key1, key2, key3, key4, key5 = obj_uuid, obj_uuid, slot_identifier, "sk", None
            cache_sk = True

        if cache_sk:
            initialize_data_dict(data, key1, key2, key3, key4, key5)
            if key4 not in data[key1][key2][key3].keys():
                data[key1][key2][key3][key4] = {}
                data[key1][key2][key3][key4][key5] = {}
            if key5 not in data[key1][key2][key3][key4].keys():
                data[key1][key2][key3][key4][key5] = {}
            data[key1][key2][key3][key4][key5][frame] = [
                k.value for k in get_sk_exported(
                    blender_obj.data.shape_keys.key_blocks)]
            cache_sk = False

        # caching driver sk meshes
        # This will avoid to have to do it again when exporting SK animation
        if blender_obj and blender_obj.type == "ARMATURE":
            sk_drivers = get_sk_drivers(obj_uuid, export_settings)
            for dr_obj in sk_drivers:
                cache_sk = False
                driver_object = export_settings['vtree'].nodes[dr_obj].blender_object
                if dr_obj not in data.keys():
                    data[dr_obj] = {}
                if blender_obj.animation_data and blender_obj.animation_data.action \
                        and blender_obj.animation_data.action_slot \
                        and export_settings['gltf_animation_mode'] in ["ACTIVE_ACTIONS", "ACTIONS", "BROADCAST"]:
                    key1, key2, key3, key4, key5 = dr_obj, obj_uuid + "_" + blender_obj.animation_data.action.name, blender_obj.animation_data.action_slot.identifier, "sk", None
                    cache_sk = True
                elif blender_obj.animation_data \
                        and export_settings['gltf_animation_mode'] in ["NLA_TRACKS"]:
                    # We can keep the input slot_identifier here, as we are caching only one object / NLA track
                    key1, key2, key3, key4, key5 = dr_obj, obj_uuid + "_" + action_name, slot_identifier, "sk", None
                    cache_sk = True
                else:
                    # slot_identifier is always None for scene export
                    key1, key2, key3, key4, key5 = dr_obj, obj_uuid + "_" + obj_uuid, slot_identifier, "sk", None
                    cache_sk = True

                if cache_sk:
                    initialize_data_dict(data, key1, key2, key3, key4, key5)
                    if export_settings['gltf_optimize_disable_viewport'] is True:
                        # Retrieve data from custom properties instead of shape keys
                        data[key1][key2][key3][key4][key5][frame] = list(
                            blender_obj['gltf_' + dr_obj])  # This include only exported SK
                    else:
                        data[key1][key2][key3][key4][key5][frame] = [
                            k.value for k in get_sk_exported(
                                driver_object.data.shape_keys.key_blocks)]
                    cache_sk = False


def light_nodetree_caching(data, action_name, slot_identifier, frame, export_settings):
    # After caching materials, caching lights, for KHR_animation_pointer
    for light in export_settings['KHR_animation_pointer']['lights'].keys():
        if len(export_settings['KHR_animation_pointer']['lights'][light]['paths']) == 0:
            continue

        blender_light = [m for m in bpy.data.lights if id(m) == light][0]
        if light not in data.keys():
            data[light] = {}

        if blender_light.node_tree and blender_light.node_tree.animation_data and blender_light.node_tree.animation_data.action \
                and blender_light.node_tree.animation_data.action_slot \
                and export_settings['gltf_animation_mode'] in ["ACTIVE_ACTIONS", "ACTIONS"]:
            key1, key2, key3, key4 = light, blender_light.node_tree.animation_data.action.name, blender_light.node_tree.animation_data.action_slot.identifier, "value"
        elif export_settings['gltf_animation_mode'] in ["NLA_TRACKS"]:
            # We can keep the input slot_identifier here, as we are caching only one object / NLA track
            key1, key2, key3, key4 = light, action_name, slot_identifier, "value"
        else:
            # case of baking materials (scene export).
            # There is no animation, so use id as key
            # slot_identifier is always None for scene export
            key1, key2, key3, key4 = light, light, slot_identifier, "value"

        if key2 not in data[key1].keys():
            data[key1][key2] = {}
            data[key1][key2][key3] = {}
            data[key1][key2][key3][key4] = {}
            for path in export_settings['KHR_animation_pointer']['lights'][light]['paths'].keys():
                data[key1][key2][key3][key4][path] = {}
        if key3 not in data[key1][key2].keys():
            data[key1][key2][key3] = {}
            data[key1][key2][key3][key4] = {}
            for path in export_settings['KHR_animation_pointer']['lights'][light]['paths'].keys():
                data[key1][key2][key3][key4][path] = {}

        for path in export_settings['KHR_animation_pointer']['lights'][light]['paths'].keys():
            val = blender_light.path_resolve(path)
            if type(val).__name__ == "float":
                data[key1][key2][key3][key4][path][frame] = val
            else:
                data[key1][key2][key3][key4][path][frame] = list(val)


def light_caching(data, action_name, slot_identifier, frame, export_settings):
    # After caching materials, caching lights, for KHR_animation_pointer
    for light in export_settings['KHR_animation_pointer']['lights'].keys():
        if len(export_settings['KHR_animation_pointer']['lights'][light]['paths']) == 0:
            continue

        blender_light = [m for m in bpy.data.lights if id(m) == light][0]
        if light not in data.keys():
            data[light] = {}

        if blender_light and blender_light.animation_data and blender_light.animation_data.action \
                and blender_light.animation_data.action_slot \
                and export_settings['gltf_animation_mode'] in ["ACTIVE_ACTIONS", "ACTIONS"]:
            key1, key2, key3, key4 = light, blender_light.animation_data.action.name, blender_light.animation_data.action_slot.identifier, "value"
        elif export_settings['gltf_animation_mode'] in ["NLA_TRACKS"]:
            key1, key2, key3, key4 = light, action_name, slot_identifier, "value"
        else:
            # case of baking materials (scene export).
            # There is no animation, so use id as key
            # slot_identifier is always None for scene export
            key1, key2, key3, key4 = light, light, slot_identifier, "value"

        if key2 not in data[key1].keys():
            data[key1][key2] = {}
            data[key1][key2][key3] = {}
            data[key1][key2][key3][key4] = {}
            for path in export_settings['KHR_animation_pointer']['lights'][light]['paths'].keys():
                data[key1][key2][key3][key4][path] = {}

        if key3 not in data[key1][key2].keys():
            data[key1][key2][key3] = {}
            data[key1][key2][key3][key4] = {}
            for path in export_settings['KHR_animation_pointer']['lights'][light]['paths'].keys():
                data[key1][key2][key3][key4][path] = {}

        for path in export_settings['KHR_animation_pointer']['lights'][light]['paths'].keys():
            # Manage special case for innerConeAngle because it requires spot_size & spot_blend
            if export_settings['KHR_animation_pointer']['lights'][light]['paths'][path]['path'] == "/extensions/KHR_lights_punctual/lights/XXX/spot.innerConeAngle":
                val = blender_light.path_resolve(path)
                val_size = blender_light.path_resolve(
                    export_settings['KHR_animation_pointer']['lights'][light]['paths'][path]['additional_path'])
                data[key1][key2][key3][key4][path][frame] = (val_size * 0.5) - ((val_size * 0.5) * val)
            else:
                # classic case
                val = blender_light.path_resolve(path)
                if type(val).__name__ == "float":
                    data[key1][key2][key3][path][frame] = val
                else:
                    # When color is coming from a node, it is 4 values (RGBA), so need to convert it to 3 values (RGB)
                    if export_settings['KHR_animation_pointer']['lights'][light]['paths'][path]['length'] == 3 and len(
                            val) == 4:
                        val = val[:3]
                    data[key1][key2][key3][key4][path][frame] = list(val)


def camera_caching(data, action_name, slot_identifier, frame, export_settings):
    # After caching lights, caching cameras, for KHR_animation_pointer
    for cam in export_settings['KHR_animation_pointer']['cameras'].keys():
        if len(export_settings['KHR_animation_pointer']['cameras'][cam]['paths']) == 0:
            continue

        blender_camera = [m for m in bpy.data.cameras if id(m) == cam][0]
        if cam not in data.keys():
            data[cam] = {}

        if blender_camera and blender_camera.animation_data and blender_camera.animation_data.action \
                and blender_camera.animation_data.action_slot \
                and export_settings['gltf_animation_mode'] in ["ACTIVE_ACTIONS", "ACTIONS"]:
            key1, key2, key3, key4 = cam, blender_camera.animation_data.action.name, blender_camera.animation_data.action_slot.identifier, "value"
        elif export_settings['gltf_animation_mode'] in ["NLA_TRACKS"]:
            # We can keep the input slot_identifier here, as we are caching only one object / NLA track
            key1, key2, key3, key4 = cam, action_name, slot_identifier, "value"
        else:
            # case of baking camera data (scene export).
            # There is no animation, so use id as key
            # No really matter for slot_identifier, as we bake all when exporting with scene export
            key1, key2, key3, key4 = cam, cam, slot_identifier, "value"

        if key2 not in data[key1].keys():
            data[key1][key2] = {}
            data[key1][key2][key3] = {}
            data[key1][key2][key3][key4] = {}
            for path in export_settings['KHR_animation_pointer']['cameras'][cam]['paths'].keys():
                data[key1][key2][key3][key4][path] = {}

        if key3 not in data[key1][key2].keys():
            data[key1][key2][key3] = {}
            data[key1][key2][key3][key4] = {}
            for path in export_settings['KHR_animation_pointer']['cameras'][cam]['paths'].keys():
                data[key1][key2][key3][ley4][path] = {}

        for path in export_settings['KHR_animation_pointer']['cameras'][cam]['paths'].keys():
            _render = bpy.context.scene.render
            width = _render.pixel_aspect_x * _render.resolution_x
            height = _render.pixel_aspect_y * _render.resolution_y
            del _render
            # Manage special case for yvof because it requires sensor_fit, aspect ratio, angle
            if export_settings['KHR_animation_pointer']['cameras'][cam]['paths'][path]['path'] == "/cameras/XXX/perspective/yfov":
                val = yvof_blender_to_gltf(blender_camera.angle, width, height, blender_camera.sensor_fit)
                data[key1][key2][key3][key4][path][frame] = val
            # Manage special case for xmag because it requires ortho_scale & scene data
            elif export_settings['KHR_animation_pointer']['cameras'][cam]['paths'][path]['path'] == "/cameras/XXX/orthographic/xmag":
                val = blender_camera.ortho_scale
                data[key1][key2][key3][key4][path][frame] = val * (width / max(width, height)) / 2.0
            # Manage special case for ymag because it requires ortho_scale  & scene data
            elif export_settings['KHR_animation_pointer']['cameras'][cam]['paths'][path]['path'] == "/cameras/XXX/orthographic/ymag":
                val = blender_camera.ortho_scale
                data[key1][key2][key3][key4][path][frame] = val * (height / max(width, height)) / 2.0
            else:
                # classic case
                val = blender_camera.path_resolve(path)
                if type(val).__name__ == "float":
                    data[key1][key2][key3][key4][path][frame] = val
                else:
                    data[key1][key2][key3][key4][path][frame] = list(val)



# SPDX-FileCopyrightText: 2019-2022 Blender Foundation
#
# SPDX-License-Identifier: GPL-2.0-or-later

import bpy

from typing import Optional
from itertools import count

from bpy.types import PoseBone

from ..utils.rig import connected_children_names
from ..utils.naming import strip_org, make_derived_name
from ..utils.bones import (put_bone, flip_bone, flip_bone_chain, is_same_position,
                           is_connected_position)
from ..utils.bones import copy_bone_position, connect_bbone_chain_handles
from ..utils.widgets_basic import create_bone_widget, create_sphere_widget
from ..utils.misc import map_list

from ..base_rig import BaseRig, stage


class SimpleChainRig(BaseRig):
    """A rig that consists of 3 connected chains of control, org and deform bones."""

    def find_org_bones(self, bone: PoseBone):
        return [bone.name] + connected_children_names(self.obj, bone.name)

    min_chain_length = 2
    bbone_segments = None

    rig_parent_bone: str  # Bone to be used as parent of the whole rig

    def initialize(self):
        if len(self.bones.org) < self.min_chain_length:
            self.raise_error(
                "Input to rig type must be a chain of {} or more bones.", self.min_chain_length)

    def parent_bones(self):
        self.rig_parent_bone = self.get_bone_parent(self.bones.org[0])

    ##############################
    # BONES

    class CtrlBones(BaseRig.CtrlBones):
        fk: list[str]                  # FK control chain

    class MchBones(BaseRig.MchBones):
        pass

    bones: BaseRig.ToplevelBones[
        list[str],
        'SimpleChainRig.CtrlBones',
        'SimpleChainRig.MchBones',
        list[str]
    ]

    ##############################
    # Control chain

    @stage.generate_bones
    def make_control_chain(self):
        self.bones.ctrl.fk = map_list(self.make_control_bone, count(0), self.bones.org)

    def make_control_bone(self, i: int, org: str):
        return self.copy_bone(org, make_derived_name(org, 'ctrl'), parent=True)

    @stage.parent_bones
    def parent_control_chain(self):
        self.parent_bone_chain(self.bones.ctrl.fk, use_connect=True)

    @stage.configure_bones
    def configure_control_chain(self):
        for args in zip(count(0), self.bones.ctrl.fk, self.bones.org):
            self.configure_control_bone(*args)

    def configure_control_bone(self, i: int, ctrl: str, org: str):
        self.copy_bone_properties(org, ctrl)

    @stage.generate_widgets
    def make_control_widgets(self):
        for args in zip(count(0), self.bones.ctrl.fk):
            self.make_control_widget(*args)

    def make_control_widget(self, i: int, ctrl: str):
        create_bone_widget(self.obj, ctrl)

    ##############################
    # ORG chain

    @stage.parent_bones
    def parent_org_chain(self):
        pass

    @stage.rig_bones
    def rig_org_chain(self):
        for args in zip(count(0), self.bones.org, self.bones.ctrl.fk):
            self.rig_org_bone(*args)

    def rig_org_bone(self, i: int, org: str, ctrl: str):
        self.make_constraint(org, 'COPY_TRANSFORMS', ctrl)

    ##############################
    # Deform chain

    @stage.generate_bones
    def make_deform_chain(self):
        self.bones.deform = map_list(self.make_deform_bone, count(0), self.bones.org)

    def make_deform_bone(self, i: int, org: str):
        name = self.copy_bone(org, make_derived_name(org, 'def'), parent=True, bbone=True)
        if self.bbone_segments:
            self.get_bone(name).bbone_segments = self.bbone_segments
        return name

    @stage.parent_bones
    def parent_deform_chain(self):
        self.parent_bone_chain(self.bones.deform, use_connect=True)

    @stage.rig_bones
    def rig_deform_chain(self):
        for args in zip(count(0), self.bones.deform, self.bones.org):
            self.rig_deform_bone(*args)

    def rig_deform_bone(self, i: int, deform: str, org: str):
        self.make_constraint(deform, 'COPY_TRANSFORMS', org)


class TweakChainRig(SimpleChainRig):
    """A rig that adds tweak controls to the triple chain."""

    ##############################
    # BONES

    class CtrlBones(SimpleChainRig.CtrlBones):
        tweak: list[str]               # Tweak control chain

    class MchBones(SimpleChainRig.MchBones):
        pass

    bones: BaseRig.ToplevelBones[
        list[str],
        'TweakChainRig.CtrlBones',
        'TweakChainRig.MchBones',
        list[str]
    ]

    ##############################
    # Tweak chain

    @stage.generate_bones
    def make_tweak_chain(self):
        orgs = self.bones.org
        self.bones.ctrl.tweak = map_list(self.make_tweak_bone, count(0), orgs + orgs[-1:])

    def make_tweak_bone(self, i: int, org: str):
        name = self.copy_bone(org, 'tweak_' + strip_org(org), parent=False, scale=0.5)

        if i == len(self.bones.org):
            put_bone(self.obj, name, self.get_bone(org).tail)

        return name

    @stage.parent_bones
    def parent_tweak_chain(self):
        ctrl = self.bones.ctrl
        for tweak, main in zip(ctrl.tweak, ctrl.fk + ctrl.fk[-1:]):
            self.set_bone_parent(tweak, main)

    @stage.configure_bones
    def configure_tweak_chain(self):
        for args in zip(count(0), self.bones.ctrl.tweak):
            self.configure_tweak_bone(*args)

    def configure_tweak_bone(self, i: int, tweak: str):
        tweak_pb = self.get_bone(tweak)
        tweak_pb.rotation_mode = 'ZXY'

        if i == len(self.bones.org):
            tweak_pb.lock_rotation_w = True
            tweak_pb.lock_rotation = (True, True, True)
            tweak_pb.lock_scale = (True, True, True)
        else:
            tweak_pb.lock_rotation_w = False
            tweak_pb.lock_rotation = (True, False, True)
            tweak_pb.lock_scale = (False, True, False)

    @stage.generate_widgets
    def make_tweak_widgets(self):
        for tweak in self.bones.ctrl.tweak:
            self.make_tweak_widget(tweak)

    def make_tweak_widget(self, tweak: str):
        create_sphere_widget(self.obj, tweak)

    ##############################
    # ORG chain

    @stage.rig_bones
    def rig_org_chain(self):
        tweaks = self.bones.ctrl.tweak
        for args in zip(count(0), self.bones.org, tweaks, tweaks[1:]):
            self.rig_org_bone(*args)

    # noinspection PyMethodOverriding
    def rig_org_bone(self, i: int, org: str, tweak: str, next_tweak: Optional[str]):
        self.make_constraint(org, 'COPY_TRANSFORMS', tweak)
        if next_tweak:
            self.make_constraint(org, 'STRETCH_TO', next_tweak, keep_axis='SWING_Y')


class ConnectingChainRig(TweakChainRig):
    """Chain rig that can attach to an end of the parent, merging bbone chains."""

    bbone_segments = 8
    use_connect_reverse = None

    use_connect_chain: bool
    connected_tweak: Optional[str]

    def initialize(self):
        super().initialize()

        self.use_connect_chain = self.params.connect_chain
        self.connected_tweak = None

        if self.use_connect_chain:
            first_org = self.bones.org[0]
            parent = self.rigify_parent

            if not isinstance(parent, SimpleChainRig):
                self.raise_error("Cannot connect to non-chain parent rig.")

            parent_orgs = parent.bones.org

            ok_reverse = is_same_position(self.obj, parent_orgs[0], first_org)
            ok_direct = is_connected_position(self.obj, parent_orgs[-1], first_org)

            if self.use_connect_reverse is None:
                self.use_connect_reverse = ok_reverse and not ok_direct

            if not (ok_reverse if self.use_connect_reverse else ok_direct):
                self.raise_error("Cannot connect chain - bone position is disjoint.")

            if isinstance(parent, ConnectingChainRig) and parent.use_connect_reverse:
                self.raise_error("Cannot connect chain - parent is reversed.")

    def prepare_bones(self):
        # Exactly match bone position to parent
        if self.use_connect_chain:
            first_bone = self.get_bone(self.bones.org[0])
            parent_orgs = self.rigify_parent.bones.org

            if self.use_connect_reverse:
                first_bone.head = self.get_bone(parent_orgs[0]).head
            else:
                first_bone.head = self.get_bone(parent_orgs[-1]).tail

    def parent_bones(self):
        # Use the parent of the shared tweak as the rig parent
        root = self.connected_tweak or self.bones.org[0]

        self.rig_parent_bone = self.get_bone_parent(root)

    ##############################
    # Control chain

    @stage.parent_bones
    def parent_control_chain(self):
        super().parent_control_chain()

        self.set_bone_parent(self.bones.ctrl.fk[0], self.rig_parent_bone)

    ##############################
    # Tweak chain

    def check_connect_tweak(self, org: str):
        """ Check if it is possible to share the last parent tweak control. """

        assert self.connected_tweak is None

        if self.use_connect_chain and isinstance(self.rigify_parent, TweakChainRig):
            # Share the last tweak bone of the parent rig
            parent_tweaks = self.rigify_parent.bones.ctrl.tweak
            index = 0 if self.use_connect_reverse else -1
            name = parent_tweaks[index]

            if not is_same_position(self.obj, name, org):
                self.raise_error("Cannot connect tweaks - position mismatch.")

            if not self.use_connect_reverse:
                copy_bone_position(self.obj, org, name, scale=0.5)

                name = self.rename_bone(name, 'tweak_' + strip_org(org))

            self.connected_tweak = parent_tweaks[index] = name

            return name
        else:
            return None

    def make_tweak_bone(self, i: int, org: str):
        if i == 0 and self.check_connect_tweak(org):
            return self.connected_tweak
        else:
            return super().make_tweak_bone(i, org)

    @stage.parent_bones
    def parent_tweak_chain(self):
        ctrl = self.bones.ctrl
        for i, tweak, main in zip(count(0), ctrl.tweak, ctrl.fk + ctrl.fk[-1:]):
            if i > 0 or not (self.connected_tweak and self.use_connect_reverse):
                self.set_bone_parent(tweak, main)

    def configure_tweak_bone(self, i: int, tweak: str):
        super().configure_tweak_bone(i, tweak)

        if self.use_connect_chain and self.use_connect_reverse and i == len(self.bones.org):
            tweak_pb = self.get_bone(tweak)
            tweak_pb.lock_rotation_w = False
            tweak_pb.lock_rotation = (True, False, True)
            tweak_pb.lock_scale = (False, True, False)

    ##############################
    # ORG chain

    @stage.parent_bones
    def parent_org_chain(self):
        if self.use_connect_chain and self.use_connect_reverse:
            flip_bone_chain(self.obj, self.bones.org)

            for org, tweak in zip(self.bones.org, self.bones.ctrl.tweak[1:]):
                self.set_bone_parent(org, tweak)

        else:
            self.set_bone_parent(self.bones.org[0], self.rig_parent_bone)

    def rig_org_bone(self, i: int, org: str, tweak: str, next_tweak: Optional[str]):
        if self.use_connect_chain and self.use_connect_reverse:
            self.make_constraint(org, 'STRETCH_TO', tweak, keep_axis='SWING_Y')
        else:
            super().rig_org_bone(i, org, tweak, next_tweak)

    ##############################
    # Deform chain

    def make_deform_bone(self, i: int, org: str):
        name = super().make_deform_bone(i, org)

        if self.use_connect_chain and self.use_connect_reverse:
            self.set_bone_parent(name, None)
            flip_bone(self.obj, name)

        return name

    @stage.parent_bones
    def parent_deform_chain(self):
        if self.use_connect_chain:
            deform = self.bones.deform
            parent_deform = self.rigify_parent.bones.deform

            if self.use_connect_reverse:
                self.set_bone_parent(deform[-1], self.bones.org[-1])
                self.parent_bone_chain(reversed(deform), use_connect=True)

                connect_bbone_chain_handles(self.obj, [deform[0], parent_deform[0]])
                return

            else:
                self.set_bone_parent(deform[0], parent_deform[-1], use_connect=True)

        super().parent_deform_chain()

    ##############################
    # Settings

    @classmethod
    def add_parameters(cls, params):
        params.connect_chain = bpy.props.BoolProperty(
            name='Connect chain',
            default=False,
            description='Connect the B-Bone chain to the parent rig'
        )

    @classmethod
    def parameters_ui(cls, layout, params):
        r = layout.row()
        r.prop(params, "connect_chain")



# SPDX-FileCopyrightText: 2009-2023 Blender Authors
#
# SPDX-License-Identifier: GPL-2.0-or-later

import bpy
from bpy.types import Panel
from bpy.app.translations import contexts as i18n_contexts
from rna_prop_ui import PropertyPanel
from bl_ui.utils import PresetPanel
from bl_ui.space_properties import PropertiesAnimationMixin


class CameraButtonsPanel:
    bl_space_type = 'PROPERTIES'
    bl_region_type = 'WINDOW'
    bl_context = "data"

    @classmethod
    def poll(cls, context):
        engine = context.engine
        return context.camera and (engine in cls.COMPAT_ENGINES)


class CAMERA_PT_presets(PresetPanel, Panel):
    bl_label = "Camera Presets"
    preset_subdir = "camera"
    preset_operator = "script.execute_preset"
    preset_add_operator = "camera.preset_add"
    COMPAT_ENGINES = {
        'BLENDER_RENDER',
        'BLENDER_EEVEE_NEXT',
        'BLENDER_WORKBENCH',
    }


class CAMERA_PT_safe_areas_presets(PresetPanel, Panel):
    bl_label = "Camera Presets"
    preset_subdir = "safe_areas"
    preset_operator = "script.execute_preset"
    preset_add_operator = "camera.safe_areas_preset_add"
    COMPAT_ENGINES = {
        'BLENDER_RENDER',
        'BLENDER_EEVEE_NEXT',
        'BLENDER_WORKBENCH',
    }


class DATA_PT_context_camera(CameraButtonsPanel, Panel):
    bl_label = ""
    bl_options = {'HIDE_HEADER'}
    COMPAT_ENGINES = {
        'BLENDER_RENDER',
        'BLENDER_EEVEE_NEXT',
        'BLENDER_WORKBENCH',
    }

    def draw(self, context):
        layout = self.layout

        ob = context.object
        cam = context.camera
        space = context.space_data

        if ob:
            layout.template_ID(ob, "data")
        elif cam:
            layout.template_ID(space, "pin_id")


class DATA_PT_lens(CameraButtonsPanel, Panel):
    bl_label = "Lens"
    bl_translation_context = i18n_contexts.id_camera
    COMPAT_ENGINES = {
        'BLENDER_RENDER',
        'BLENDER_EEVEE_NEXT',
        'BLENDER_WORKBENCH',
    }

    def draw(self, context):
        layout = self.layout
        layout.use_property_split = True

        cam = context.camera

        layout.prop(cam, "type")

        col = layout.column()
        col.separator()

        if cam.type == 'PERSP':
            if cam.lens_unit == 'MILLIMETERS':
                col.prop(cam, "lens")
            elif cam.lens_unit == 'FOV':
                col.prop(cam, "angle")
            col.prop(cam, "lens_unit")

        elif cam.type == 'ORTHO':
            col.prop(cam, "ortho_scale")

        elif cam.type == 'PANO':
            engine = context.engine
            if engine == 'CYCLES':
                col.prop(cam, "panorama_type")
                if cam.panorama_type == 'FISHEYE_EQUIDISTANT':
                    col.prop(cam, "fisheye_fov")
                elif cam.panorama_type == 'FISHEYE_EQUISOLID':
                    col.prop(cam, "fisheye_lens", text="Lens")
                    col.prop(cam, "fisheye_fov")
                elif cam.panorama_type == 'EQUIRECTANGULAR':
                    sub = col.column(align=True)
                    sub.prop(cam, "latitude_min", text="Latitude Min")
                    sub.prop(cam, "latitude_max", text="Max")
                    sub = col.column(align=True)
                    sub.prop(cam, "longitude_min", text="Longitude Min")
                    sub.prop(cam, "longitude_max", text="Max")
                elif cam.panorama_type == 'FISHEYE_LENS_POLYNOMIAL':
                    col.prop(cam, "fisheye_fov")
                    col.prop(cam, "fisheye_polynomial_k0", text="K0")
                    col.prop(cam, "fisheye_polynomial_k1", text="K1")
                    col.prop(cam, "fisheye_polynomial_k2", text="K2")
                    col.prop(cam, "fisheye_polynomial_k3", text="K3")
                    col.prop(cam, "fisheye_polynomial_k4", text="K4")
                elif cam.panorama_type == 'CENTRAL_CYLINDRICAL':
                    sub = col.column(align=True)
                    sub.prop(cam, "central_cylindrical_range_v_min", text="Height Min")
                    sub.prop(cam, "central_cylindrical_range_v_max", text="Max")
                    sub = col.column(align=True)
                    sub.prop(cam, "central_cylindrical_range_u_min", text="Longitude Min")
                    sub.prop(cam, "central_cylindrical_range_u_max", text="Max")
                    sub = col.column(align=True)
                    sub.prop(cam, "central_cylindrical_radius", text="Cylinder Radius")

            elif engine in {'BLENDER_RENDER', 'BLENDER_EEVEE_NEXT', 'BLENDER_WORKBENCH'}:
                if cam.lens_unit == 'MILLIMETERS':
                    col.prop(cam, "lens")
                elif cam.lens_unit == 'FOV':
                    col.prop(cam, "angle")
                col.prop(cam, "lens_unit")

        elif cam.type == 'CUSTOM':
            engine = context.engine
            if engine == 'CYCLES':
                sub = col.row()
                sub.prop(cam, "custom_mode", text=" ", expand=True)

                sub = col.row(align=True)
                if cam.custom_mode == 'EXTERNAL':
                    sub.prop(cam, "custom_filepath", text=" ")
                else:
                    sub.prop(cam, "custom_shader", text=" ")
                sub.operator("object.camera_custom_update", icon='FILE_REFRESH', text="")

        col = layout.column()
        col.separator()

        sub = col.column(align=True)
        sub.prop(cam, "shift_x", text="Shift X")
        sub.prop(cam, "shift_y", text="Y")

        col.separator()
        sub = col.column(align=True)
        sub.prop(cam, "clip_start", text="Clip Start")
        sub.prop(cam, "clip_end", text="End", text_ctxt=i18n_contexts.id_camera)


class DATA_PT_camera_stereoscopy(CameraButtonsPanel, Panel):
    bl_label = "Stereoscopy"
    COMPAT_ENGINES = {
        'BLENDER_RENDER',
        'BLENDER_EEVEE_NEXT',
        'BLENDER_WORKBENCH',
    }

    @classmethod
    def poll(cls, context):
        render = context.scene.render
        return (super().poll(context) and render.use_multiview and render.views_format == 'STEREO_3D')

    def draw(self, context):
        layout = self.layout
        layout.use_property_split = True

        render = context.scene.render
        st = context.camera.stereo
        cam = context.camera

        is_spherical_stereo = cam.type != 'ORTHO' and render.use_spherical_stereo
        use_spherical_stereo = is_spherical_stereo and st.use_spherical_stereo

        layout.prop(st, "convergence_mode")

        col = layout.column()
        sub = col.column()
        sub.active = st.convergence_mode != 'PARALLEL'
        sub.prop(st, "convergence_distance")

        col.prop(st, "interocular_distance")

        if is_spherical_stereo:
            col.separator()
            col.prop(st, "use_spherical_stereo")
            sub = col.column()
            sub.active = st.use_spherical_stereo
            sub.prop(st, "use_pole_merge")

            sub = col.column(align=True)
            sub.active = st.use_pole_merge
            sub.prop(st, "pole_merge_angle_from", text="Pole Merge Angle Start")
            sub.prop(st, "pole_merge_angle_to", text="End")

        col = layout.column()
        col.active = not use_spherical_stereo
        col.separator()
        col.prop(st, "pivot")


class DATA_PT_camera(CameraButtonsPanel, Panel):
    bl_label = "Camera"
    bl_options = {'DEFAULT_CLOSED'}
    COMPAT_ENGINES = {
        'BLENDER_RENDER',
        'BLENDER_EEVEE_NEXT',
        'BLENDER_WORKBENCH',
    }

    def draw_header_preset(self, _context):
        CAMERA_PT_presets.draw_panel_header(self.layout)

    def draw(self, context):
        layout = self.layout

        cam = context.camera

        layout.use_property_split = True

        col = layout.column()
        col.prop(cam, "sensor_fit")

        if cam.sensor_fit == 'AUTO':
            col.prop(cam, "sensor_width", text="Size")
        else:
            sub = col.column(align=True)
            sub.active = cam.sensor_fit == 'HORIZONTAL'
            sub.prop(cam, "sensor_width", text="Width")

            sub = col.column(align=True)
            sub.active = cam.sensor_fit == 'VERTICAL'
            sub.prop(cam, "sensor_height", text="Height")


class DATA_PT_camera_dof(CameraButtonsPanel, Panel):
    bl_label = "Depth of Field"
    bl_options = {'DEFAULT_CLOSED'}
    COMPAT_ENGINES = {
        'BLENDER_EEVEE_NEXT',
        'BLENDER_WORKBENCH',
    }

    def draw_header(self, context):
        cam = context.camera
        dof = cam.dof
        self.layout.prop(dof, "use_dof", text="")

    def draw(self, context):
        layout = self.layout
        layout.use_property_split = True

        cam = context.camera
        dof = cam.dof
        layout.active = dof.use_dof

        col = layout.column()
        col.prop(dof, "focus_object", text="Focus on Object")
        if dof.focus_object and dof.focus_object.type == 'ARMATURE':
            col.prop_search(dof, "focus_subtarget", dof.focus_object.data, "bones", text="Focus on Bone")

        sub = col.column()
        sub.active = (dof.focus_object is None)
        row = sub.row(align=True)
        row.prop(dof, "focus_distance", text="Focus Distance")
        row.operator(
            "ui.eyedropper_depth",
            icon='EYEDROPPER',
            text="",
        ).prop_data_path = "scene.camera.data.dof.focus_distance"


class DATA_PT_camera_dof_aperture(CameraButtonsPanel, Panel):
    bl_label = "Aperture"
    bl_parent_id = "DATA_PT_camera_dof"
    COMPAT_ENGINES = {
        'BLENDER_EEVEE_NEXT',
        'BLENDER_WORKBENCH',
    }

    def draw(self, context):
        layout = self.layout
        layout.use_property_split = True

        cam = context.camera
        dof = cam.dof
        layout.active = dof.use_dof

        flow = layout.grid_flow(row_major=True, columns=0, even_columns=True, even_rows=False, align=False)

        col = flow.column()
        col.prop(dof, "aperture_fstop")

        col = flow.column()
        col.prop(dof, "aperture_blades")
        col.prop(dof, "aperture_rotation")
        col.prop(dof, "aperture_ratio")


class DATA_PT_camera_background_image(CameraButtonsPanel, Panel):
    bl_label = "Background Images"
    bl_options = {'DEFAULT_CLOSED'}
    COMPAT_ENGINES = {
        'BLENDER_RENDER',
        'BLENDER_EEVEE_NEXT',
        'BLENDER_WORKBENCH',
    }

    def draw_header(self, context):
        cam = context.camera

        self.layout.prop(cam, "show_background_images", text="")

    def draw(self, context):
        layout = self.layout
        layout.use_property_split = True
        layout.use_property_decorate = False

        cam = context.camera
        use_multiview = context.scene.render.use_multiview

        col = layout.column()
        col.operator("view3d.camera_background_image_add", text="Add Image")

        for i, bg in enumerate(cam.background_images):
            layout.active = cam.show_background_images
            box = layout.box()
            row = box.row(align=True)
            row.prop(bg, "show_expanded", text="", emboss=False)
            if bg.source == 'IMAGE' and bg.image:
                row.prop(bg.image, "name", text="", emboss=False)
            elif bg.source == 'MOVIE_CLIP' and bg.clip:
                row.prop(bg.clip, "name", text="", emboss=False)
            elif bg.source and bg.use_camera_clip:
                row.label(text="Active Clip")
            else:
                row.label(text="Not Set")

            row.prop(
                bg,
                "show_background_image",
                text="",
                emboss=False,
                icon='RESTRICT_VIEW_OFF' if bg.show_background_image else 'RESTRICT_VIEW_ON',
            )

            row.operator("view3d.camera_background_image_remove", text="", emboss=False, icon='X').index = i

            if bg.show_expanded:
                row = box.row()
                row.prop(bg, "source", expand=True)

                has_bg = False
                if bg.source == 'IMAGE':
                    row = box.row()
                    row.template_ID(bg, "image", open="image.open")
                    if bg.image is not None:
                        box.template_image(bg, "image", bg.image_user, compact=True)
                        has_bg = True

                        if use_multiview:
                            box.prop(bg.image, "use_multiview")

                            column = box.column()
                            column.active = bg.image.use_multiview

                            column.label(text="Views Format:")
                            column.row().prop(bg.image, "views_format", expand=True)

                            sub = column.box()
                            sub.active = bg.image.views_format == 'STEREO_3D'
                            sub.template_image_stereo_3d(bg.image.stereo_3d_format)

                elif bg.source == 'MOVIE_CLIP':
                    box.prop(bg, "use_camera_clip", text="Active Clip")

                    column = box.column()
                    column.active = not bg.use_camera_clip
                    column.template_ID(bg, "clip", open="clip.open")

                    if bg.clip:
                        column.template_movieclip(bg, "clip", compact=True)

                    if bg.use_camera_clip or bg.clip:
                        has_bg = True

                    column = box.column()
                    column.active = has_bg
                    column.prop(bg.clip_user, "use_render_undistorted")
                    column.prop(bg.clip_user, "proxy_render_size")

                if has_bg:
                    col = box.column()
                    if bg.image is not None:
                        col.prop(bg.image, "use_view_as_render")
                    col.prop(bg, "alpha")
                    col.row().prop(bg, "display_depth", expand=True)

                    col.row().prop(bg, "frame_method", expand=True)

                    row = box.row()
                    row.prop(bg, "offset")

                    col = box.column()
                    col.prop(bg, "rotation")
                    col.prop(bg, "scale")

                    col = box.column(heading="Flip", heading_ctxt=i18n_contexts.id_image)
                    col.prop(bg, "use_flip_x", text="X")
                    col.prop(bg, "use_flip_y", text="Y")


class DATA_PT_camera_display(CameraButtonsPanel, Panel):
    bl_label = "Viewport Display"
    bl_options = {'DEFAULT_CLOSED'}
    COMPAT_ENGINES = {
        'BLENDER_RENDER',
        'BLENDER_EEVEE_NEXT',
        'BLENDER_WORKBENCH',
    }

    def draw(self, context):
        layout = self.layout
        layout.use_property_split = True

        cam = context.camera

        col = layout.column(align=True)

        col.prop(cam, "display_size", text="Size")

        col = layout.column(heading="Show")
        col.prop(cam, "show_limits", text="Limits")
        col.prop(cam, "show_mist", text="Mist")
        col.prop(cam, "show_sensor", text="Sensor")
        col.prop(cam, "show_name", text="Name")

        col = layout.column(align=False, heading="Passepartout")
        col.use_property_decorate = False
        row = col.row(align=True)
        sub = row.row(align=True)
        sub.prop(cam, "show_passepartout", text="")
        sub = sub.row(align=True)
        sub.active = cam.show_passepartout
        sub.prop(cam, "passepartout_alpha", text="")
        row.prop_decorator(cam, "passepartout_alpha")


class DATA_PT_camera_display_composition_guides(CameraButtonsPanel, Panel):
    bl_label = "Composition Guides"
    bl_parent_id = "DATA_PT_camera_display"
    bl_options = {'DEFAULT_CLOSED'}
    COMPAT_ENGINES = {
        'BLENDER_RENDER',
        'BLENDER_EEVEE_NEXT',
        'BLENDER_WORKBENCH',
    }

    def draw(self, context):
        layout = self.layout
        layout.use_property_split = True

        cam = context.camera

        layout.prop(cam, "show_composition_thirds")

        col = layout.column(heading="Center", align=True)
        col.prop(cam, "show_composition_center")
        col.prop(cam, "show_composition_center_diagonal", text="Diagonal")

        col = layout.column(heading="Golden", align=True)
        col.prop(cam, "show_composition_golden", text="Ratio")
        col.prop(cam, "show_composition_golden_tria_a", text="Triangle A")
        col.prop(cam, "show_composition_golden_tria_b", text="Triangle B")

        col = layout.column(heading="Harmony", align=True)
        col.prop(cam, "show_composition_harmony_tri_a", text="Triangle A")
        col.prop(cam, "show_composition_harmony_tri_b", text="Triangle B")


class DATA_PT_camera_safe_areas(CameraButtonsPanel, Panel):
    bl_label = "Safe Areas"
    bl_options = {'DEFAULT_CLOSED'}
    COMPAT_ENGINES = {
        'BLENDER_RENDER',
        'BLENDER_EEVEE_NEXT',
        'BLENDER_WORKBENCH',
    }

    def draw_header(self, context):
        cam = context.camera

        self.layout.prop(cam, "show_safe_areas", text="")

    def draw_header_preset(self, _context):
        CAMERA_PT_safe_areas_presets.draw_panel_header(self.layout)

    def draw(self, context):
        layout = self.layout
        safe_data = context.scene.safe_areas
        camera = context.camera

        layout.use_property_split = True

        layout.active = camera.show_safe_areas

        col = layout.column()

        sub = col.column()
        sub.prop(safe_data, "title", slider=True)
        sub.prop(safe_data, "action", slider=True)


class DATA_PT_camera_safe_areas_center_cut(CameraButtonsPanel, Panel):
    bl_label = "Center-Cut Safe Areas"
    bl_parent_id = "DATA_PT_camera_safe_areas"
    bl_options = {'DEFAULT_CLOSED'}
    COMPAT_ENGINES = {
        'BLENDER_RENDER',
        'BLENDER_EEVEE_NEXT',
        'BLENDER_WORKBENCH',
    }

    def draw_header(self, context):
        cam = context.camera

        layout = self.layout
        layout.active = cam.show_safe_areas
        layout.prop(cam, "show_safe_center", text="")

    def draw(self, context):
        layout = self.layout
        safe_data = context.scene.safe_areas
        camera = context.camera

        layout.use_property_split = True

        layout.active = camera.show_safe_areas and camera.show_safe_center

        col = layout.column()
        col.prop(safe_data, "title_center", slider=True)
        col.prop(safe_data, "action_center", slider=True)


class DATA_PT_camera_animation(CameraButtonsPanel, PropertiesAnimationMixin, PropertyPanel, Panel):
    COMPAT_ENGINES = {
        'BLENDER_RENDER',
        'BLENDER_EEVEE_NEXT',
        'BLENDER_WORKBENCH',
    }
    _animated_id_context_property = "camera"


class DATA_PT_custom_props_camera(CameraButtonsPanel, PropertyPanel, Panel):
    COMPAT_ENGINES = {
        'BLENDER_RENDER',
        'BLENDER_EEVEE_NEXT',
        'BLENDER_WORKBENCH',
    }
    _context_path = "object.data"
    _property_type = bpy.types.Camera


classes = (
    CAMERA_PT_presets,
    CAMERA_PT_safe_areas_presets,
    DATA_PT_context_camera,
    DATA_PT_lens,
    DATA_PT_camera_dof,
    DATA_PT_camera_dof_aperture,
    DATA_PT_camera,
    DATA_PT_camera_stereoscopy,
    DATA_PT_camera_safe_areas,
    DATA_PT_camera_safe_areas_center_cut,
    DATA_PT_camera_background_image,
    DATA_PT_camera_display,
    DATA_PT_camera_display_composition_guides,
    DATA_PT_camera_animation,
    DATA_PT_custom_props_camera,
)

if __name__ == "__main__":  # only for live edit.
    from bpy.utils import register_class
    for cls in classes:
        register_class(cls)



# SPDX-FileCopyrightText: 2017-2022 Blender Foundation
#
# SPDX-License-Identifier: GPL-2.0-or-later

import bpy

from ...utils.rig import connected_children_names
from ...utils.layers import ControlLayersOption
from ...utils.bones import BoneUtilityMixin, flip_bone_chain

from ...base_generate import SubstitutionRig

from . import basic_spine, basic_tail, super_head


class Rig(SubstitutionRig, BoneUtilityMixin):
    """Compatibility proxy for the monolithic super_spine rig that splits it into parts."""

    def substitute(self):
        params_copy = self.params_copy
        orgs = [self.base_bone] + connected_children_names(self.obj, self.base_bone)

        # Split the bone list according to the settings
        spine_orgs = orgs
        head_orgs = None
        tail_orgs = None

        pivot_pos = self.params.pivot_pos

        if self.params.use_head:
            neck_pos = self.params.neck_pos
            if neck_pos <= pivot_pos:
                self.raise_error("Neck cannot be below or the same as pivot.")
            if neck_pos >= len(orgs):
                self.raise_error("Neck is too short.")

            spine_orgs = orgs[0: neck_pos - 1]
            head_orgs = orgs[neck_pos - 1:]

        if self.params.use_tail:
            tail_pos = self.params.tail_pos
            if tail_pos < 2:
                self.raise_error("Tail is too short.")
            if tail_pos >= pivot_pos:
                self.raise_error("Tail cannot be above or the same as pivot.")

            tail_orgs = list(reversed(spine_orgs[0: tail_pos]))
            spine_orgs = spine_orgs[tail_pos:]
            pivot_pos -= tail_pos

        # Split the bone chain and flip the tail
        if head_orgs or tail_orgs:
            bpy.ops.object.mode_set(mode='EDIT')

            if spine_orgs[0] != orgs[0]:
                self.set_bone_parent(spine_orgs[0], self.get_bone_parent(orgs[0]))

            if head_orgs:
                self.get_bone(head_orgs[0]).use_connect = False

            if tail_orgs:
                flip_bone_chain(self.obj, reversed(tail_orgs))
                self.set_bone_parent(tail_orgs[0], spine_orgs[0])

            bpy.ops.object.mode_set(mode='OBJECT')

        # Create the parts
        self.assign_params(spine_orgs[0], params_copy, pivot_pos=pivot_pos, make_fk_controls=False)

        result = [self.instantiate_rig(basic_spine.Rig, spine_orgs[0])]

        if tail_orgs:
            self.assign_params(tail_orgs[0], params_copy, connect_chain=True)

            result += [self.instantiate_rig(basic_tail.Rig, tail_orgs[0])]

        if head_orgs:
            self.assign_params(head_orgs[0], params_copy, connect_chain=True)

            result += [self.instantiate_rig(super_head.Rig, head_orgs[0])]

        return result


def add_parameters(params):
    basic_spine.Rig.add_parameters(params)
    basic_tail.Rig.add_parameters(params)
    super_head.Rig.add_parameters(params)

    params.neck_pos = bpy.props.IntProperty(
        name="Neck Position",
        default=6,
        min=0,
        description='Neck start position'
    )

    params.tail_pos = bpy.props.IntProperty(
        name="Tail Position",
        default=2,
        min=2,
        description='Where the tail starts'
    )

    params.use_tail = bpy.props.BoolProperty(
        name="Use Tail",
        default=False,
        description='Create tail bones'
    )

    params.use_head = bpy.props.BoolProperty(
        name="Use Head",
        default=True,
        description='Create head and neck bones'
    )


def parameters_ui(layout, params):
    """ Create the ui for the rig parameters."""

    layout.label(text="Note: this combined rig is deprecated.", icon='INFO')

    r = layout.row(align=True)
    r.prop(params, "use_head", toggle=True, text="Head")
    r.prop(params, "use_tail", toggle=True, text="Tail")

    r = layout.row()
    r.prop(params, "neck_pos")
    r.enabled = params.use_head

    r = layout.row()
    r.prop(params, "pivot_pos")

    r = layout.row()
    r.prop(params, "tail_pos")
    r.enabled = params.use_tail

    r = layout.row()
    col = r.column(align=True)
    row = col.row(align=True)
    for i, axis in enumerate(['x', 'y', 'z']):
        row.prop(params, "copy_rotation_axes", index=i, toggle=True, text=axis)
    r.enabled = params.use_tail

    ControlLayersOption.TWEAK.parameters_ui(layout, params)


def create_sample(obj):
    bones = basic_spine.create_sample(obj)
    basic_tail.create_sample(obj, parent=bones['spine'])
    super_head.create_sample(obj, parent=bones['spine.003'])
